import mdtraj
import argparse
import math
import numpy as np
import pickle
import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'large'
mpl.rcParams['xtick.labelsize'] = 'large'
mpl.rcParams['ytick.labelsize'] = 'large'
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
from itertools import product
import pandas as pd
import ray
import os

parser = argparse.ArgumentParser()
parser.add_argument('--elec_type', type = str)
parser.add_argument('--ss_type', type = str)
args = parser.parse_args()

elec_type = args.elec_type
ss_type = args.ss_type

#print(f"elec_type: {elec_type:10}, ss_type: {ss_type}")

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
protein_names = protein_names.iloc[:, 0].tolist()
weight_decay_list = [5e-7, 1e-6, 2e-6]

weight_decay = 1e-6

with open(f"./output/plots/rmsd_and_rg.pkl", 'rb') as file_handle:
    data = pickle.load(file_handle)
    rmsd_md = data['rmsd_md']
    rg_md = data['rg_md']
    rmsd_cg = data['rmsd_cg']
    rg_cg = data['rg_cg']

for name in protein_names:
    rmsd_cg[weight_decay][name] = rmsd_cg[weight_decay][name][10000:]
    rg_cg[weight_decay][name] = rg_cg[weight_decay][name][10000:]    

#### rmsd ####
rmsd_max = {'CLN025': 0.80, '2JOF': 1.28, '1FME': 1.53, '2F4K': 1.53,
         'GTT': 1.53, 'NTL9': 1.30, '2WAV': 2.03, 'PRB': 2.33,
         'UVF': 2.03, 'NuG2': 1.63, 'A3D': 2.53, 'lambda': 1.73}

rg_max = {'CLN025': 0.93, '2JOF': 1.53, '1FME': 2.03, '2F4K': 1.83,
         'GTT': 1.53, 'NTL9': 1.23, '2WAV': 2.03, 'PRB': 2.33,
         'UVF': 2.03, 'NuG2': 1.63, 'A3D': 2.63, 'lambda': 1.73}

stride = 800
fig = plt.figure(figsize = (6.4*2, 4.8*3))

for name in protein_names:
    print(name)
    idx_protein = protein_names.index(name)
    if idx_protein % 2 == 1:
        idx_subplot = idx_protein * 2
    else:
        idx_subplot = idx_protein * 2 + 1
        
    axes = fig.add_subplot(12, 2, idx_subplot)
    plt.plot(rmsd_md[name][::stride], ms = 3, color = 'C1', linewidth = 1)    
    s = int(len(rmsd_cg[weight_decay][name]) / len(rmsd_md[name]) * stride)
    plt.plot(rmsd_cg[weight_decay][name][::s], ms = 3, color = 'C0', linewidth = 1)    
    plt.ylim(0, rmsd_max[name])
    axes.set_xticks([])
    #plt.ylabel("RMSD (nm)")
    
    axes = fig.add_subplot(12, 2, idx_subplot + 2)    
    plt.plot(rg_md[name][::stride], ms = 3, color = 'C1', linewidth = 1)
    s = int(len(rg_cg[weight_decay][name]) / len(rg_md[name]) * stride)
    plt.plot(rg_cg[weight_decay][name][::s], ms = 3, color = 'C0', linewidth = 1)    
    plt.ylim(None, rg_max[name])
    axes.set_xticks([])
    #plt.ylabel("Rg (nm)")    

plt.tight_layout()    
plt.savefig(f"./output/figures/rmsd_rg_traj_all.eps")
plt.savefig(f"./output/figures/rmsd_rg_traj_all.png")    

exit()

ncols = 4
nrows = 3

fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
for name in protein_names:
    print(name)
    axes = fig.add_subplot(nrows, ncols, protein_names.index(name) + 1)
    bins = np.linspace(0, rmsd_max[name], 40)
    axes.hist(rmsd_md[name], bins = bins, alpha = 0.5, label = 'All atom', density = True, color = 'C1', log = True)
    axes.hist(rmsd_cg[weight_decay][name], bins = bins, alpha = 0.5, label = 'CG', density = True, color = 'C0', log = True)
    if name == 'CLN025':
        axes.legend()
    plt.xlim(-0.1, rmsd_max[name])
os.makedirs(f"./output/figures", exist_ok = True)    
plt.savefig(f"./output/figures/rmsd_hist_log.pdf")

exit()

fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
for name in protein_names:
    print(name)
    axes = fig.add_subplot(nrows, ncols, protein_names.index(name) + 1)
    bins = np.linspace(0, rmsd_max[name], 40)
    axes.hist(rmsd_md[name], bins = bins, alpha = 0.5, label = 'All atom', density = True, color = 'C1')
    axes.hist(rmsd_cg[weight_decay][name], bins = bins, alpha = 0.5, label = 'CG', density = True, color = 'C0')
    if name == 'CLN025':
        axes.legend()
    plt.xlim(-0.1, rmsd_max[name])
os.makedirs(f"./output/figures", exist_ok = True)    
plt.savefig(f"./output/figures/rmsd_hist.eps")





stride = 200
fig = plt.figure(figsize = (6.4*4, 4.8*4))
for name in protein_names:
    print(name)
    axes = fig.add_subplot(12, 1, protein_names.index(name) + 1)
    plt.plot(rmsd_md[name][::stride], '.', label = 'All atom', ms = 3, color = 'C1')
    
    s = int(len(rmsd_cg[weight_decay][name]) / len(rmsd_md[name]) * stride)
    
    plt.plot(rmsd_cg[weight_decay][name][::s], '.', label = 'CG', ms = 3, color = 'C0')    
    plt.ylim(0, rmsd_max[name])
    axes.set_xticks([])
    
plt.savefig(f"./output/figures/rmsd_traj.eps")
plt.savefig(f"./output/figures/rmsd_traj.png")    

#### rg ####
ncols = 4
nrows = 3
fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
rg_max = {'CLN025': 0.93, '2JOF': 1.53, '1FME': 2.03, '2F4K': 1.83,
         'GTT': 1.53, 'NTL9': 1.23, '2WAV': 2.03, 'PRB': 2.33,
         'UVF': 2.03, 'NuG2': 1.63, 'A3D': 2.63, 'lambda': 1.73}
for name in protein_names:
    print(name)
    axes = fig.add_subplot(nrows, ncols, protein_names.index(name) + 1)
    rg_min = min(rg_md[name].min(), rg_cg[weight_decay][name].min())
    bins = np.linspace(rg_min, rg_max[name], 40)
    axes.hist(rg_md[name], bins = bins, alpha = 0.5, label = 'All atom', density = True, color = 'C1')
    axes.hist(rg_cg[weight_decay][name], bins = bins, alpha = 0.5, label = 'CG', density = True, color = 'C0')
    if name == 'CLN025':
        axes.legend()
    #plt.xlim(-0.1, rg_max[name])
os.makedirs(f"./output/figures", exist_ok = True)    
plt.savefig(f"./output/figures/rg_hist.eps")

fig = plt.figure(figsize = (6.4*4, 4.8*4))
for name in protein_names:
    print(name)
    axes = fig.add_subplot(12, 1, protein_names.index(name) + 1)
    plt.plot(rg_md[name][::stride], '.', label = 'All atom', ms = 3, color = 'C1')

    s = int(len(rg_cg[weight_decay][name]) / len(rg_md[name]) * stride)
    
    plt.plot(rg_cg[weight_decay][name][::s], '.', label = 'CG', ms = 3, color = 'C0')    
    plt.ylim(None, rg_max[name])
    axes.set_xticks([])
    
plt.savefig(f"./output/figures/rg_traj.eps")
plt.savefig(f"./output/figures/rg_traj.png")

#fig = plt.figure(figsize = (6.4*3, 4.8*3))
fig = plt.figure(figsize = (6.4*3, 4.8*3))
idx_axes = 1
for name in ['2JOF', 'A3D']:
    axes = fig.add_subplot(4, 1, idx_axes)
    plt.plot(rmsd_md[name][::stride], label = 'All atom', ms = 3, color = 'C1', linewidth = 1)
    s = int(len(rmsd_cg[weight_decay][name]) / len(rmsd_md[name]) * stride)
    plt.plot(rmsd_cg[weight_decay][name][::s], label = 'CG', ms = 3, color = 'C0', linewidth = 1)
    plt.ylim(0, rmsd_max[name])
    axes.set_xticks([])
    idx_axes += 1

    axes = fig.add_subplot(4, 1, idx_axes)    
    plt.plot(rg_md[name][::stride], label = 'All atom', ms = 3, color = 'C1', linewidth = 1)
    s = int(len(rg_cg[weight_decay][name]) / len(rg_md[name]) * stride)
    plt.plot(rg_cg[weight_decay][name][::s], label = 'CG', ms = 3, color = 'C0', linewidth = 1)    
    plt.ylim(None, rg_max[name])
    axes.set_xticks([])
    idx_axes += 1
    
plt.savefig(f"./output/figures/rmsd_and_rg_traj.eps")


exit()



import mdtraj
import argparse
import math
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
from itertools import product
import pandas as pd
import ray

parser = argparse.ArgumentParser()
parser.add_argument('--elec_type', type = str)
parser.add_argument('--ss_type', type = str)
args = parser.parse_args()

elec_type = args.elec_type
ss_type = args.ss_type

print(f"elec_type: {elec_type:10}, ss_type: {ss_type}")

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
weight_decay_list = [5e-7,
                     1e-6, 2e-6]
protein_names = protein_names.iloc[:, 0].tolist()

psf = {}
traj_ref = {}
for name in protein_names:
    print(name)
    psf[name] = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")    
    traj_ref[name] = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf[name])

print("******* MD *********")    
rmsd_md = {}
rg_md = {}
for name in protein_names:
    print(name)
    traj_md = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", psf[name])
    rmsd_md[name] = mdtraj.rmsd(traj_md, traj_ref[name])
    rg_md[name] = mdtraj.compute_rg(traj_md)

print("******* CG *********")        
rmsd_cg = {}
rg_cg = {}
for weight_decay in weight_decay_list:
    rmsd_cg[weight_decay] = {}
    rg_cg[weight_decay] = {}    
    for name in protein_names:
        print(weight_decay, name)
        traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.dcd",
                                  psf[name], stride = 4)    
        rmsd_cg[weight_decay][name] = mdtraj.rmsd(traj_cg, traj_ref[name])
        rg_cg[weight_decay][name] = mdtraj.compute_rg(traj_cg)

with open(f"./output/plots/rmsd_and_rg.pkl", 'wb') as file_handle:
    pickle.dump({'rmsd_md': rmsd_md,
                 'rg_md': rg_md,
                 'rmsd_cg': rmsd_cg,
                 'rg_cg': rg_cg}, file_handle)
    
exit()



import pickle
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
mpl.rc('font', size = 16)
mpl.rc('axes', titlesize = 'large', labelsize = 'large')
mpl.rc('xtick', labelsize = 'large')
mpl.rc('ytick', labelsize = 'large')
from matplotlib.backends.backend_pdf import PdfPages
import os
import numpy as np
import pandas as pd
from sys import exit
from matplotlib.pyplot import cm
import argparse
import sys
import torch

with open(f"./output/common/LJ_rmin.pkl", 'rb') as file_handle:
    r_min_dict = pickle.load(file_handle)
aa_pairs = list(r_min_dict.keys())
aa_pairs.sort()

AA_info = pd.read_csv("./info/amino_acids_with_learned_sigmas.csv",
                      index_col = 'name')
AA_names = list(AA_info.index)

AA_name_dict = {'G': 'GLY', 'P': 'PRO', 'A': 'ALA', 'V': 'VAL',
                'L': 'LEU', 'I': 'ILE', 'M': 'MET', 'C': 'CYS',
                'F': 'PHE', 'Y': 'TYR', 'W': 'TRP', 'H': 'HIS',
                'K': 'LYS', 'R': 'ARG', 'Q': 'GLN', 'N': 'ASN',
                'E': 'GLU', 'D': 'ASP', 'S': 'SER', 'T': 'THR'}

NUM_COLORS = 20

name = "CLN025"

weight_decay_list = [1e-10, 1e-9, 1e-8,
                     1e-7, 1e-6, 1e-5,
                     1e-4, 1e-3]
weight_decay = 1e-6

pdf = PdfPages(f"./output/plots/LJ_potential.pdf")

for weight_decay in weight_decay_list:
    print(weight_decay)
    
    data = torch.load(f"./output/{name}/FF/FF_weight_decay_{weight_decay:.3E}.pt",
                      map_location=torch.device('cpu'))
    FF = data['FF']
    bonded_ff, lj_ff, rmsd_ff = FF['bonded'], FF['lj'], FF['rmsd']

    ncols = 5
    nrows = 4
    fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
    fig.suptitle(f'LJ_weight_decay:{weight_decay:.3E}')
    for aa1 in AA_names:
        print(aa1)
        plt.subplot(nrows, ncols, AA_names.index(aa1) + 1)
        colors = iter([cm.tab20(i) for i in range(20)])
        for aa2 in AA_names:
            aa = [aa1, aa2]
            aa.sort()
            if (aa[0], aa[1]) in aa_pairs:
                idx = aa_pairs.index((aa[0], aa[1]))
                r = np.linspace(lj_ff['min'], lj_ff['max'], lj_ff['U'].shape[-1])
                plt.plot()
                plt.plot(r, lj_ff['U'][idx],
                        linewidth = 2.0,
                        label = aa2,
                        color = next(colors))
        plt.ylim(-10, 10)
        plt.xlim(0, lj_ff['max'])
        #plt.legend(ncol = 2, loc = 'right')
        plt.title(aa1)
    pdf.savefig()

pdf.close()

exit()

ncol = 6
nrow = len(aa_pairs)//ncol + 1

fig = plt.figure(np.random.randint(1000), figsize = (6.4*ncol, 4.8*nrow))
fig.clf()
for k in range(len(aa_pairs)):
    print(k)
    plt.subplot(nrow,ncol,k+1)
    plt.plot(LJ_parameters[aa_pairs[k]]['r_over_range'], LJ_parameters[aa_pairs[k]]['U'], linewidth = 4.0)
    plt.ylim(-10, 10)
    plt.xlim(0, 1.5)
    aa1, aa2 = aa_pairs[k]
    plt.title(f"{aa1}-{aa2}")
plt.tight_layout()    
os.makedirs("./output/plots", exist_ok = True)
plt.savefig(f"./output/plots/LJ_potential_split_weight_decay_{weight_decay:.3E}.pdf")

AA_names = [AA_name_dict[n] for n in 'RHKDESTNQCGPAVILMFYW']
        
E_min = np.zeros((len(AA_names), len(AA_names)))
for i in range(len(AA_names)):
    for j in range(len(AA_names)):
        aa1, aa2 = AA_names[i], AA_names[j]
        aa = [aa1, aa2]
        aa.sort()
        if (aa[0], aa[1]) in aa_pairs:
            E_min[i,j] = np.min(LJ_parameters[(aa[0], aa[1])]['U'])
        else:
            E_min[i,j] = None

fig = plt.figure(np.random.randint(1000), figsize = (6.4*2, 4.8*2))
fig.clf()
ax = plt.subplot()
im, cbar = heatmap(E_min, AA_names, AA_names, ax = ax, origin = 'lower')
fig.tight_layout()
plt.savefig(f"./output/plots/LJ_heatmap_weight_decay_{weight_decay:.3E}.pdf")

AA_names = [AA_name_dict[n] for n in 'MRGDPNKSAFVLTYHEQIWC']
E_min = np.zeros((len(AA_names), len(AA_names)))
for i in range(len(AA_names)):
    for j in range(len(AA_names)):
        aa1, aa2 = AA_names[i], AA_names[j]
        aa = [aa1, aa2]
        aa.sort()
        if (aa[0], aa[1]) in aa_pairs:
            E_min[i,j] = np.min(LJ_parameters[(aa[0], aa[1])]['U'])
        else:
            E_min[i,j] = None

fig = plt.figure(np.random.randint(1000), figsize = (6.4*2, 4.8*2))
fig.clf()
ax = plt.subplot()
im, cbar = heatmap(E_min, AA_names, AA_names, ax = ax)
fig.tight_layout()
plt.savefig(f"./output/plots/LJ_heatmap_MOFF_weight_decay_{weight_decay:.3E}.pdf")

exit()


# fig = plt.figure(np.random.randint(1000), figsize = (6.4*2, 4.8*2))
# ax = plt.subplot()
# im = ax.imshow(E_min)
# ax.set_xticks(np.arange(E_min.shape[0]))
# ax.set_yticks(np.arange(E_min.shape[1]))
# ax.set_xticklabels(AA_names)
# ax.set_yticklabels(AA_names)
# plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
#          rotation_mode="anchor")
# plt.legend()
# fig.tight_layout()
# plt.savefig(f"./output/plots/LJ_heatmap_weight_decay_{weight_decay:.3E}.pdf")

exit()



ncol = 6
nrow = len(aa_pairs)//ncol + 1

fig = plt.figure(np.random.randint(1000), figsize = (6.4*ncol, 4.8*nrow))
fig.clf()
for k in range(len(aa_pairs)):
    print(k)
    plt.subplot(nrow,ncol,k+1)
    plt.plot(LJ_parameters[aa_pairs[k]]['r_over_range'], LJ_parameters[aa_pairs[k]]['U'], linewidth = 2.0)
    plt.ylim(-10, 20)

os.makedirs("./output/plots", exist_ok = True)
plt.savefig(f"./output/plots/LJ_potential_weight_decay_{weight_decay:.3E}.pdf")




#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=learn_im
#SBATCH --time=3:00:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-1
#SBATCH --mem=100G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/learn_im_%a.txt

import numpy as np
import torch
torch.set_default_dtype(torch.float64)
import pandas as pd
from sys import exit
import simtk.unit as unit
from scipy import optimize
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import pickle
import os
import argparse
import mdtraj
import simtk.unit as unit
import math
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
from MMFlow import utils
from CLCG.utils.splines import *
from CLCG.utils.CL import *
import pandas as pd

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
    coor_transformer = pickle.load(file_handle)
    
data = torch.load(f"./output/{name}/ic_md.pt")
ic = data['ic']
ic.double()

ic.reference_particle_3_angle.clamp_(0.0, math.pi)
ic.angle.clamp_(0.0, math.pi)
ic.dihedral.clamp_(-math.pi, math.pi)

bonded_parameters = {}
bonded_parameters['reference_particle_2_bond'] = {
    'b0': torch.mean(ic.reference_particle_2_bond).item(),
    'kb': 1/torch.var(ic.reference_particle_2_bond).item()
}

bonded_parameters['reference_particle_3_bond'] = {
    'b0': torch.mean(ic.reference_particle_3_bond).item(),
    'kb': 1/torch.var(ic.reference_particle_3_bond).item()
}

bonded_parameters['bond'] = {
    'b0': torch.mean(ic.bond, 0),
    'kb': 1/torch.var(ic.bond, 0)
}

angle_knots = torch.linspace(0, math.pi, 20)[1:-1]
angle_boundary_knots = torch.tensor([0.0, math.pi])

dihedral_knots = torch.linspace(-math.pi, math.pi, 20)[1:-1]
dihedral_boundary_knots = torch.tensor([-math.pi, math.pi])

torch.save(
    {'angle_knots': angle_knots,
     'angle_boundary_knots': angle_boundary_knots,
     'dihedral_knots': dihedral_knots,
     'dihedral_boundary_knots': dihedral_boundary_knots},
    f"./output/{name}/knots_for_angle_and_dihedral.pt"
)
    
angle_md = torch.cat([ic.reference_particle_3_angle[:, None], ic.angle], dim = -1)
angle_parameters = []

pdf = PdfPages(f"./output/{name}/angle_hist_from_md_and_im.pdf")

for j in range(angle_md.shape[-1]):
    basis_md = bs(angle_md[:,j], angle_knots, angle_boundary_knots)
    
    angle_noise = torch.distributions.Uniform(0, math.pi).sample((angle_md.shape[0],))
    basis_noise = bs(angle_noise, angle_knots, angle_boundary_knots)
    
    log_q_md = basis_md.new_ones(basis_md.shape[0])*math.log(1./np.pi) 
    log_q_noise = basis_noise.new_ones(basis_noise.shape[0])*math.log(1./np.pi)

    alphas, F = contrastive_learning(log_q_noise, log_q_md,
                                     basis_noise, basis_md)
    
    a_over_the_range = torch.linspace(0.0, math.pi, 1000)    
    basis_over_the_range = bs(a_over_the_range, angle_knots, angle_boundary_knots)
    U = torch.matmul(basis_over_the_range, alphas)

    fig = plt.figure(0)
    fig.clf()
    axes = plt.hist(angle_md[:,j].numpy(), bins = 50, range = (0.0, np.pi),
                    density = True,
                    log = False, label = 'md')
    p = np.exp(-(U - F))
    p = p*axes[0].max()/p.max()

    plt.plot(a_over_the_range, p, '-', label = 'cl', linewidth = 4)
    plt.legend()
    pdf.savefig()

    angle_parameters.append({'alphas': alphas,
                             'F': F,
                             'U': U,
                             'a_over_the_range': a_over_the_range})    
pdf.close()

bonded_parameters['reference_particle_3_angle'] = angle_parameters[0]
bonded_parameters['angle'] = angle_parameters[1:]

dihedral_md = ic.dihedral
dihedral_parameters = []

os.makedirs(f"./output/{name}", exist_ok = True)
pdf = PdfPages(f"./output/{name}/dihedral_hist_from_md_and_cl.pdf")

for j in range(dihedral_md.shape[-1]):
    basis_md = pbs(dihedral_md[:,j], dihedral_knots, dihedral_boundary_knots)
    
    dihedral_noise = torch.distributions.Uniform(-math.pi, math.pi).sample((dihedral_md.shape[0],))
    basis_noise = pbs(dihedral_noise, dihedral_knots, dihedral_boundary_knots)
    
    log_q_md = basis_md.new_ones(basis_md.shape[0])*math.log(1./(2*np.pi)) 
    log_q_noise = basis_noise.new_ones(basis_noise.shape[0])*math.log(1./(2*np.pi))
    
    alphas, F = contrastive_learning(log_q_noise, log_q_md,
                                     basis_noise, basis_md)

    d_over_the_range = torch.linspace(-math.pi, math.pi, 1000)    
    basis_over_the_range = pbs(d_over_the_range, dihedral_knots, dihedral_boundary_knots)
    U = torch.matmul(basis_over_the_range, alphas)
    
    fig = plt.figure(0)
    fig.clf()
    axes = plt.hist(dihedral_md[:,j].numpy(), bins = 50, range = (-np.pi, np.pi),
                    density = True,
                    log = False, label = 'md')
    p = np.exp(-(U - F))
    p = p*axes[0].max()/p.max()

    plt.plot(d_over_the_range, p, '-', label = 'cl', linewidth = 4)
    plt.legend()
    pdf.savefig()

    dihedral_parameters.append({'alphas': alphas,
                                'F': F,
                                'U': U,
                                'd_over_the_range': d_over_the_range})
pdf.close()

bonded_parameters['dihedral'] = dihedral_parameters

os.makedirs(f"./output/{name}", exist_ok = True)
with open(f"./output/{name}/bonded_parameters_im.pkl", 'wb') as file_handle:
    pickle.dump(bonded_parameters,
                file_handle)
exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=basis_angle_dihedral
#SBATCH --time=01:00:00
#SBATCH --partition=xeon-p8
#SBATCH --cpus-per-task=4
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --mem=100G
#SBATCH --array=0-11
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/basis_angle_dihedral_%a.txt

import numpy as np
import torch
torch.set_default_dtype(torch.float64)
import pandas as pd
from sys import exit
import simtk.unit as unit
from scipy import optimize
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import pickle
import os
import argparse
import mdtraj
import simtk.unit as unit
import math
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
from MMFlow import utils
from CLCG.utils.splines import *
from CLCG.utils.CL import *
from scipy.sparse import csr_matrix
import pandas as pd
import os

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
    coor_transformer = pickle.load(file_handle)
    
data = torch.load(f"./output/{name}/ic_md.pt")
ic_md = data['ic']

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']
traj_imus = []
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj_imus.append(traj)
traj_imus = mdtraj.join(traj_imus)
stride = size // 10
traj_imus = traj_imus[::stride]

xyz_imus = traj_imus.xyz.astype(np.float64)
ic_imus, ic_logabsdet = coor_transformer.compute_internal_coordinate_from_xyz(torch.from_numpy(xyz_imus))
torch.save({'ic': ic_imus, 'ic_logabsdet': ic_logabsdet},
           f"./output/{name}/ic_imus.pt")

ic_md.double()
ic_imus.double()

ic_md.reference_particle_3_angle.clamp_(0.0, math.pi)
ic_md.angle.clamp_(0.0, math.pi)
ic_md.dihedral.clamp_(-math.pi, math.pi)

ic_imus.reference_particle_3_angle.clamp_(0.0, math.pi)
ic_imus.angle.clamp_(0.0, math.pi)
ic_imus.dihedral.clamp_(-math.pi, math.pi)

data = torch.load(f"./output/{name}/knots_for_angle_and_dihedral.pt")
angle_knots = data['angle_knots']
angle_boundary_knots = data['angle_boundary_knots']

dihedral_knots = data['dihedral_knots']
dihedral_boundary_knots = data['dihedral_boundary_knots']

basis_md = {}
basis_md['reference_particle_3_angle'] = bs(ic_md.reference_particle_3_angle, angle_knots, angle_boundary_knots)
basis_md['angle'] = torch.cat(
    [bs(ic_md.angle[:,j], angle_knots, angle_boundary_knots) for j in range(ic_md.angle.shape[1])],
    dim = -1)

basis_md['dihedral'] = torch.cat(
    [pbs(ic_md.dihedral[:,j],
         dihedral_knots,
         dihedral_boundary_knots)
     for j in range(ic_md.dihedral.shape[1])],
    dim = -1)

basis_imus = {}
basis_imus['reference_particle_3_angle'] = bs(ic_imus.reference_particle_3_angle,
                                              angle_knots,
                                              angle_boundary_knots)
basis_imus['angle'] = torch.cat(
    [bs(ic_imus.angle[:,j],
        angle_knots,
        angle_boundary_knots)
     for j in range(ic_imus.angle.shape[1])],
    dim = -1)
basis_imus['dihedral'] = torch.cat(
    [pbs(ic_imus.dihedral[:,j],
         dihedral_knots,
         dihedral_boundary_knots)
     for j in range(ic_imus.dihedral.shape[1])],
    dim = -1)

# for k in basis_md.keys():
#     basis_md[k] = basis_md[k].to_sparse()
#     basis_imus[k] = basis_imus[k].to_sparse()        

basis_grid = {}
basis_grid['angle_min'] = 0.0
basis_grid['angle_max'] = math.pi
angle_grid = torch.linspace(basis_grid['angle_min'], basis_grid['angle_max'], 1000)
basis_grid['angle_basis'] = bs(angle_grid,
                               angle_knots,
                               angle_boundary_knots)

basis_grid['dihedral_min'] = -math.pi
basis_grid['dihedral_max'] = math.pi
dihedral_grid = torch.linspace(basis_grid['dihedral_min'], basis_grid['dihedral_max'], 1000)
basis_grid['dihedral_basis'] = pbs(dihedral_grid,
                                   dihedral_knots,
                                   dihedral_boundary_knots)

torch.save({'md': basis_md, 'imus': basis_imus, 'grid': basis_grid},
           f'./output/{name}/basis_angle_and_dihedrals.pt')

exit()    
    
    




import mdtraj
import argparse
import math
import numpy as np
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
from itertools import product
import pandas as pd
import ray
import os
import scipy.cluster.hierarchy
from scipy.spatial.distance import squareform

parser = argparse.ArgumentParser()
parser.add_argument('--elec_type', type = str)
parser.add_argument('--ss_type', type = str)
args = parser.parse_args()

elec_type = args.elec_type
ss_type = args.ss_type

print(f"elec_type: {elec_type:10}, ss_type: {ss_type}")

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
weight_decay_list = [5e-7,
                     1e-6, 2e-6]
#protein_names = protein_names.iloc[[1,-2], 0].tolist()
protein_names = protein_names.iloc[:, 0].tolist()

psf = {}
traj_ref = {}
for name in protein_names:
    print(name)
    psf[name] = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")    
    traj_ref[name] = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf[name])
    traj_ref[name].save_xyz(f"./output/figures/structure_with_lowest_rmsd/{name}_reference.xyz")

print("******* CG *********")        
rmsd_cg = {}
weight_decay = 1e-6

for name in protein_names:
    traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.dcd",
                              psf[name], stride = 4)
    traj_cg = traj_cg[10000:]

    N = 5_000
    stride = int(traj_cg.n_frames / N)
    traj_sub = traj_cg[::stride]
    traj_sub = traj_sub[0:N]

    distances = np.empty((traj_sub.n_frames, traj_sub.n_frames))
    for i in range(traj_sub.n_frames):
        distances[i] = mdtraj.rmsd(traj_sub, traj_sub, i)    
        # if (i + 1) % 200 == 0:
        #     print(i)

    #print('Max pairwise rmsd: %f nm' % np.max(distances))
    reduced_distances = squareform(distances, checks=False)

    linkage = scipy.cluster.hierarchy.linkage(reduced_distances, method='average')
    n_cluster = 50
    cluster_label = scipy.cluster.hierarchy.fcluster(linkage, t = n_cluster, criterion = 'maxclust')
    cluster_size = []
    for i in range(1, n_cluster+1):
        cluster_size.append(np.sum(cluster_label == i))

    popular_cluster_label = np.argmax(cluster_size) + 1
    flag = cluster_label == popular_cluster_label
    traj_sub = traj_sub[flag]

    dist = np.mean(distances[flag,:][:,flag], -1)
    idx = np.argmin(dist)

    os.makedirs(f"./output/figures/structure_cluster/", exist_ok = True)
    traj_sub.superpose(traj_ref[name])
    traj_sub[idx].save_xyz(f"./output/figures/structure_with_lowest_rmsd/{name}.xyz")

    rmsd = mdtraj.rmsd(traj_sub[idx], traj_ref[name])
    print(weight_decay, name, rmsd)
    
exit()



import mdtraj
import numpy as np
import pickle
import pandas as pd
from itertools import product
from sys import exit

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
weight_decay_list = [1e-8, 2e-8, 5e-8, 1e-7, 2e-7, 5e-7, 1e-6]
protein_names = protein_names.iloc[0:3, 0].tolist()

for name, weight_decay in product(protein_names, weight_decay_list):
    print(f"name: {name}, weight_decay: {weight_decay}", flush = True)
    
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    with open(f"./output/{name}/TRE/weight_decay_{weight_decay:.3E}/log_temperatures.pkl", 'rb') as file_handle:
        log_t = pickle.load(file_handle)

    size = log_t.shape[0]
    T = log_t.min()
    traj_list = []
    for rank in range(size):
        traj = mdtraj.load_dcd(f"./output/{name}/TRE/weight_decay_{weight_decay:.3E}/traj_{rank}.dcd", psf)
        traj = traj[log_t[rank] == T]
        traj_list.append(traj)
    traj = mdtraj.join(traj_list)
    traj.save_dcd(f"./output/{name}/TRE/weight_decay_{weight_decay:.3E}/traj.dcd")
    



import torch
import torch.distributions as distributions
from pccg.splines import bs
from pccg.CL import contrastive_learning
import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np

#### define the Muller potential
def compute_Muller_potential(scale, x):
    A = (-200.0, -100.0, -170.0, 15.0)
    beta = (0.0, 0.0, 11.0, 0.6)
    alpha_gamma = (
        x.new_tensor([-1.0, -10.0]),
        x.new_tensor([-1.0, -10.0]),
        x.new_tensor([-6.5, -6.5]),
        x.new_tensor([0.7, 0.7]),
    )

    ab = (
        x.new_tensor([1.0, 0.0]),
        x.new_tensor([0.0, 0.5]),
        x.new_tensor([-0.5, 1.5]),
        x.new_tensor([-1.0, 1.0]),
    )

    U = 0
    for i in range(4):
        diff = x - ab[i]
        U = U + A[i] * torch.exp(
            torch.sum(alpha_gamma[i] * diff**2, -1) + beta[i] * torch.prod(diff, -1)
        )

    U = scale * U
    return U


def generate_grid(x1_min, x1_max, x2_min, x2_max, size=100):
    x1 = torch.linspace(x1_min, x1_max, size)
    x2 = torch.linspace(x2_min, x2_max, size)
    grid_x1, grid_x2 = torch.meshgrid(x1, x2, indexing="ij")
    grid = torch.stack([grid_x1, grid_x2], dim=-1)
    x = grid.reshape((-1, 2))
    return x


x1_min, x1_max = -1.5, 1.0
x2_min, x2_max = -0.5, 2.0

x = generate_grid(x1_min, x1_max, x2_min, x2_max)
fig, axes = plt.subplots()
scale = 0.05
U = compute_Muller_potential(scale, x)
U = U.reshape(100, 100)
U[U > 9] = 9
U = U.T
plt.contourf(
    U,
    levels=np.linspace(-9, 9, 19),
    extent=(x1_min, x1_max, x2_min, x2_max),
    cmap=cm.viridis_r,
)
plt.xlabel(r"$x_1$", fontsize=24)
plt.ylabel(r"$x_2$", fontsize=24)
plt.colorbar()
axes.set_aspect("equal")
plt.tight_layout()
plt.savefig("./output/potential.png")

## draw samples from the Müller potential using temperature replica exchange
## Monte Carlo sampling
############################################################################

num_reps = 10  # number of replicas
scales = torch.linspace(0.0, scale, num_reps)

num_steps = 110000
x_record = []
accept_rate = 0
x = torch.stack(
    (
        x1_min + torch.rand(num_reps) * (x1_max - x1_min),
        x2_min + torch.rand(num_reps) * (x2_max - x2_min),
    ),
    dim=-1,
)
energy = compute_Muller_potential(1.0, x)

for k in range(num_steps):
    if (k + 1) % 10000 == 0:
        print("steps: {} out of {} total steps".format(k, num_steps))

    ## sampling within each replica
    delta_x = torch.normal(0, 1, size=(num_reps, 2)) * 0.3
    x_p = x + delta_x
    energy_p = compute_Muller_potential(1.0, x_p)

    ## accept based on energy
    accept_prop = torch.exp(-scales * (energy_p - energy))
    accept_flag = torch.rand(num_reps) < accept_prop

    ## considering the bounding effects
    accept_flag = (
        accept_flag
        & torch.all(x_p > x_p.new_tensor([x1_min, x2_min]), -1)
        & torch.all(x_p < x_p.new_tensor([x1_max, x2_max]), -1)
    )

    x_p[~accept_flag] = x[~accept_flag]
    energy_p[~accept_flag] = energy[~accept_flag]
    x = x_p
    energy = energy_p

    ## calculate overall accept rate
    accept_rate = accept_rate + (accept_flag.float() - accept_rate) / (k + 1)

    ## exchange
    if k % 10 == 0:
        for i in range(1, num_reps):
            accept_prop = torch.exp(
                (scales[i] - scales[i - 1]) * (energy[i] - energy[i - 1])
            )
            accept_flag = torch.rand(1) < accept_prop
            if accept_flag.item():
                tmp = x[i].clone()
                x[i] = x[i - 1]
                x[i - 1] = tmp

                tmp = energy[i].clone()
                energy[i] = energy[i - 1]
                energy[i - 1] = tmp

        if k >= 10000:
            x_record.append(x.clone().numpy())

x_record = np.array(x_record)
x_samples = x_record[:, -1, :]

#### plot samples
fig = plt.figure()
fig.clf()
plt.plot(x_samples[:, 0], x_samples[:, 1], ".", alpha=0.5)
plt.xlim((x1_min, x1_max))
plt.ylim((x2_min, x2_max))
plt.xlabel(r"$x_1$", fontsize=24)
plt.ylabel(r"$x_2$", fontsize=24)
axes.set_aspect("equal")
plt.tight_layout()
plt.savefig("./output/samples.png")

####
xp = torch.from_numpy(x_samples)
num_samples_p = xp.shape[0]

## samples from q
num_samples_q = num_samples_p

q_dist = distributions.Independent(
    distributions.Uniform(
        low=torch.tensor([x1_min, x2_min]), high=torch.tensor([x1_max, x2_max])
    ),
    1,
)
xq = q_dist.sample((num_samples_q,))

fig, axes = plt.subplots()
plt.plot(xp[::10, 0], xp[::10, 1], ".", label="data", markersize=6)
plt.plot(xq[::10, 0], xq[::10, 1], ".", label="noise", markersize=6)
plt.xlabel(r"$x_1$", fontsize=24)
plt.ylabel(r"$x_2$", fontsize=24)
plt.tick_params(
    which="both",
    bottom=False,
    top=False,
    right=False,
    left=False,
    labelbottom=False,
    labelleft=False,
)
plt.tight_layout()
axes.set_aspect("equal")
plt.savefig("./output/data_and_noise_samples.png")


x1_knots = torch.linspace(x1_min, x1_max, steps=10)[1:-1]
x2_knots = torch.linspace(x2_min, x2_max, steps=10)[1:-1]

x1_boundary_knots = torch.tensor([x1_min, x1_max])
x2_boundary_knots = torch.tensor([x2_min, x2_max])


def compute_basis(x, x1_knots, x2_knots, x1_boundary_knots, x2_boundary_knots):
    x1_basis = bs(x[:, 0], x1_knots, x1_boundary_knots)
    x2_basis = bs(x[:, 1], x2_knots, x2_boundary_knots)
    x_basis = x1_basis[:, :, None] * x2_basis[:, None, :]
    x_basis = x_basis.reshape([x_basis.shape[0], -1])
    return x_basis


xp_basis = compute_basis(xp, x1_knots, x2_knots, x1_boundary_knots, x2_boundary_knots)
xq_basis = compute_basis(xq, x1_knots, x2_knots, x1_boundary_knots, x2_boundary_knots)

log_q_noise = q_dist.log_prob(xq)
log_q_data = q_dist.log_prob(xp)

theta, F = contrastive_learning(
    log_q_noise,
    log_q_data,
    xq_basis,
    xp_basis,
    options={"disp": True, "gtol": 1e-6, "ftol": 1e-12},
)

x_grid = generate_grid(x1_min, x1_max, x2_min, x2_max, size=100)
x_grid_basis = compute_basis(
    x_grid, x1_knots, x2_knots, x1_boundary_knots, x2_boundary_knots
)
up = torch.matmul(x_grid_basis, theta)
up = up.reshape(100, 100)
up = up.T.numpy()

up = up - up.min() + -7.3296
fig, axes = plt.subplots()
plt.contourf(
    up,
    levels=np.linspace(-9, 9, 19),
    extent=(x1_min, x1_max, x2_min, x2_max),
    cmap=cm.viridis_r,
)
plt.xlabel(r"$x_1$", fontsize=24)
plt.ylabel(r"$x_2$", fontsize=24)
plt.tick_params(
    which="both",
    bottom=False,
    top=False,
    right=False,
    left=False,
    labelbottom=False,
    labelleft=False,
)
plt.colorbar()
plt.tight_layout()
axes.set_aspect("equal")
plt.savefig("./output/learned_up.png")


#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=CL
#SBATCH --time=10:00:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --mem=100G
#SBATCH --array=0-3
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/cl_%a.txt

import argparse
import numpy as np
import torch
torch.set_default_dtype(torch.float64)
import pickle
from scipy import optimize
from sys import exit
import mdtraj
import pandas as pd
import openmm.unit as unit
from FastMBAR import *
import sys
import MMFlow
from MMFlow import utils
import pyarrow.feather as feather
import os
import math
import torch.nn as nn
from functools import reduce
from scipy.sparse import coo_matrix, csr_matrix
import itertools
import argparse
import ray
from collections import OrderedDict

parser = argparse.ArgumentParser()
parser.add_argument("--weight_decay", type = float)
parser.add_argument("--elec_type", type = str)
parser.add_argument("--ss_type", type = str)
#parser.add_argument("--bounds_flag", type = bool)
args = parser.parse_args()

weight_decay = args.weight_decay
ss_type = args.ss_type
elec_type = args.elec_type
#bounds_flag = args.bounds_flag
print(f"weight_decay: {weight_decay:.3E}")
print(f"elec_type: {elec_type}")
print(f"ss_type: {ss_type}")
#print(f"bounds_flag: {bounds_flag}")

protein_names = pd.read_csv("./info/protein_names.txt",
                            comment = "#",
                            header = None)
protein_names = protein_names.iloc[:, 0].tolist()

@ray.remote
class CL_Actor(torch.nn.Module):
    def __init__(self, name):
        #super(CL_Actor, self).__init__()
        torch.set_default_dtype(torch.float64)        
        super().__init__()        
        self.name = name
        
        protein_info = pd.read_csv(
            "./info/protein_temperature_and_ionic_strength.txt",
            index_col = 'name',
            comment = '#')
        self.T = protein_info.loc[self.name, 'temperature']
        
        ic_imus = torch.load(f"./output/{name}/ic_imus.pt")['ic']
        ic_md = torch.load(f"./output/{name}/ic_md.pt")['ic']

        self.N_imus = len(ic_imus)
        self.N_md = len(ic_md)
        self.N = self.N_imus + self.N_md
        
        self.target = torch.cat([torch.zeros(self.N_imus), torch.ones(self.N_md)])
        
        bond_imus = torch.cat(
            (ic_imus.reference_particle_2_bond[:, None],
             ic_imus.reference_particle_3_bond[:, None],
             ic_imus.bond), dim = -1)
        bond_md = torch.cat(
            (ic_md.reference_particle_2_bond[:, None],
             ic_md.reference_particle_3_bond[:, None],
             ic_md.bond), dim = -1)

        bond = torch.cat([bond_imus, bond_md])
        del ic_imus, ic_md, bond_imus, bond_md

        data = torch.load(f'./output/{name}/basis_angle_and_dihedrals.pt')
        self.rp3angle = torch.cat(
            [data['imus']['reference_particle_3_angle'],
             data['md']['reference_particle_3_angle']])
        self.angle = torch.cat(
            [data['imus']['angle'],
             data['md']['angle']])
        self.dihedral = torch.cat(
            [data['imus']['dihedral'],
             data['md']['dihedral']])
        
        self.angle_grid = data['grid']['angle_basis']
        self.angle_min = data['grid']['angle_min']        
        self.angle_max = data['grid']['angle_max']
        
        self.dihedral_grid = data['grid']['dihedral_basis']
        self.dihedral_min = data['grid']['dihedral_min']        
        self.dihedral_max = data['grid']['dihedral_max']
        del data

        with open(f"./output/{name}/bonded_parameters_im.pkl", 'rb') as file_handle:
            bonded_parameters = pickle.load(file_handle)
            self.bonded_parameters = bonded_parameters
            
        b0 = torch.cat([torch.tensor([bonded_parameters['reference_particle_2_bond']['b0']]),
                        torch.tensor([bonded_parameters['reference_particle_3_bond']['b0']]),
                        bonded_parameters['bond']['b0']])
        kb = torch.cat([torch.tensor([bonded_parameters['reference_particle_2_bond']['kb']]),
                        torch.tensor([bonded_parameters['reference_particle_3_bond']['kb']]),
                        bonded_parameters['bond']['kb']])
        bond_energy = torch.sum(0.5*kb*(bond - b0)**2, -1)
        self.register_buffer('bond_energy', bond_energy, persistent = False)
        
        self.c3a = nn.Parameter(bonded_parameters['reference_particle_3_angle']['alphas'])
        self.ca = nn.Parameter(torch.cat([p['alphas'] for p in bonded_parameters['angle']]))
        self.cd = nn.Parameter(torch.cat([p['alphas'] for p in bonded_parameters['dihedral']]))

        data = torch.load(f"./output/{name}/basis_lj.pt")
        self.lj = torch.cat([data['imus'], data['md']])
                
        data = torch.load(f"./output/common/basis_lj_grid.pt")
        self.lj_grid = data['basis']
        self.lj_min = data['min']
        self.lj_max = data['max']

        lj_omega = data['omega']
        self.register_buffer('lj_omega', lj_omega, persistent = False)
        self.n_lj, self.nb_lj = lj_omega.shape[0], lj_omega.shape[1]
        clj = torch.zeros((self.n_lj, self.nb_lj))
        clj[:, 0] = 5.0
        clj = clj.view(-1)
        self.clj = nn.Parameter(clj)        
        #self.clj = nn.Parameter(torch.zeros(self.lj.shape[-1]))
        del data

        if os.path.exists(f"./output/{name}/basis_rmsd_{ss_type}.pt"):
            data = torch.load(f"./output/{name}/basis_rmsd_{ss_type}.pt")
            rmsd_imus = torch.cat([d['imus'] for d in data], dim = -1)
            rmsd_md = torch.cat([d['md'] for d in data], dim = -1)
            self.rmsd = torch.cat([rmsd_imus, rmsd_md])
            self.rmsd_grid = torch.stack([d['grid']['basis'] for d in data])
            self.rmsd_max = torch.tensor([d['grid']['max'] for d in data])            
            self.rmsd_min = torch.tensor([d['grid']['min'] for d in data])

            self.n_rmsd = len(data)
            self.nb_rmsd = self.rmsd_grid.shape[-1]
            del data, rmsd_imus, rmsd_md
            self.crmsd = nn.Parameter(torch.zeros(self.rmsd.shape[-1]))
            
        else:
            self.rmsd = None
            self.rmsd_min = None
            self.rmsd_max = None            
            self.n_rmsd = 0
            self.nb_rmsd = 0
            self.crmsd = nn.Parameter(torch.tensor([]))

        self.rmsd = None
        self.rmsd_min = None
        self.rmsd_max = None            
        self.n_rmsd = 0
        self.nb_rmsd = 0
        self.crmsd = nn.Parameter(torch.tensor([]))

            
        data = torch.load(f"./output/{name}/elec_energy_{elec_type}.pt")
        elec_energy = torch.cat([data['imus'], data['md']])
        self.register_buffer('elec_energy', elec_energy, persistent = False)
        del data

        data = torch.load(f"./output/{name}/log_q_imus.pt")
        self.log_q = torch.cat([data['imus'], data['md']])

        self._numel_cache = None

        ## initialize self.F
        with torch.no_grad():
            u_p = self.bond_energy + self.elec_energy + \
                  self.rp3angle.mv(self.c3a) + self.angle.mv(self.ca) + \
                  self.dihedral.mv(self.cd)
        u_p = u_p.numpy()
        u_q = -self.log_q.numpy()

        energy_matrix = np.vstack([u_q, u_p])
        num_conf = np.array([self.N_imus, self.N_md], dtype = np.float64)
        fastmbar = FastMBAR(energy_matrix, num_conf, verbose = True)
        
        self.F = nn.Parameter(torch.tensor([fastmbar.F[-1]]))
        
        # ## Because pytorch 1.10 has limited support on sparse csr matrix,
        # ## we use scipy.sparse.csr_matrix for now
        # self.rp3angle = self._torch_coo_to_scipy_csr(self.rp3angle)        
        # self.angle = self._torch_coo_to_scipy_csr(self.angle)
        # self.dihedral = self._torch_coo_to_scipy_csr(self.dihedral)
        # self.lj = self._torch_coo_to_scipy_csr(self.lj)                

        print(f"finish initializing {name}", flush = True)
        
    def _numel(self):
        if self._numel_cache is None:
            self._numel_cache = reduce(lambda total, p: total + p.numel(), self.parameters(), 0)
        return self._numel_cache

    def get_named_parameter_numel(self):
        res = OrderedDict()
        for name, p in self.named_parameters():
            res[name] = p.numel()
        return res
    
    def _gather_flat_grad(self):
        views = []
        for p in self.parameters():
            if p.grad is None:
                view =  p.new(p.numel()).zero_()
            else:
                view = p.grad.view(-1)
            views.append(view)
        return torch.cat(views, 0)

    def _gather_flat_param(self):
        views = []
        for p in self.parameters():
            view = p.data.view(-1)
            views.append(view)
        return torch.cat(views, 0)
    
    def _set_param(self, update):
        offset = 0
        for p in self.parameters():
            numel = p.numel()
            p.data.copy_(update[offset:offset + numel].view_as(p))
            offset += numel
        assert offset == self._numel()                

    def _torch_coo_to_scipy_csr(self, s):
        s = s.coalesce()
        row, col = s.indices().tolist()
        data = s.values().numpy()
        shape = list(s.shape)
        return coo_matrix((data, (row, col)), shape = shape).tocsr()

    def _scipy_csr_to_torch_coo(self, s):
        s = s.tocoo()
        s = torch.sparse_coo_tensor(np.array([s.row, s.col]), s.data, size = s.shape)
        return s

    def make_parameter_bounds(self):
        bounds = []
        for name, param in self.named_parameters():
            if name == 'clj':
                for i,j in itertools.product(range(self.n_lj), range(self.nb_lj)):
                    if j == 0:
                        bounds.append([0, None])
                    else:
                        bounds.append([None, None])
            else:
                bounds += [[None, None] for _ in range(param.numel())]
        assert(len(bounds) == self._numel())
        return bounds

    def compute_loss_and_grad(self, x, weight_decay):
        if type(x) == np.ndarray:
            x = torch.from_numpy(x)
            
        device = self.ca.device
        x = x.to(device)
        
        self._set_param(x)
        self.zero_grad()

        self.rp3angle = self.rp3angle.to(device)
        self.angle = self.angle.to(device)
        self.dihedral = self.dihedral.to(device)
        self.lj = self.lj.to(device)        
        self.log_q = self.log_q.to(device)
        self.target = self.target.to(device)

        kbT = unit.BOLTZMANN_CONSTANT_kB * self.T * unit.kelvin * unit.AVOGADRO_CONSTANT_NA
        kbT_in_kJ_per_mole = kbT.value_in_unit(unit.kilojoule_per_mole)

        ## note all energy terms except LJ are in reduced units
        ## at the corresponding temperature.
        ## LJ energy term has the unit of kcal/mol
        u = self.bond_energy + self.elec_energy + \
            self.rp3angle.mv(self.c3a) + self.angle.mv(self.ca) + \
            self.dihedral.mv(self.cd) + self.lj.mv(self.clj)/kbT_in_kJ_per_mole
        
        if self.rmsd is not None:
            self.rmsd = self.rmsd.to(device)                    
            u = u + self.rmsd.mv(self.crmsd)
                
        log_p = - (u - self.F) + u.new_tensor([self.N_md/self.N_imus])
        logit = log_p - self.log_q
        loss = torch.nn.functional.binary_cross_entropy_with_logits(
            logit, self.target, reduction = 'mean')
        loss.backward()
        loss = loss.cpu().item()
        
        clj = self.clj.view(self.n_lj, self.nb_lj)
        loss_wc = 0.5*weight_decay*torch.sum(clj[:,None,:].matmul(self.lj_omega).matmul(clj[:,:,None]))
        loss_wc.backward()
        
        loss += loss_wc.cpu().item()
        grad = self._gather_flat_grad()

        return loss, grad.cpu().numpy()
    
    def compute_loss_and_grad_in_batch(self, x, weight_decay, batch_size):
        if type(x) == np.ndarray:
            x = torch.from_numpy(x)
            
        device = self.ca.device
        x = x.to(device)
        
        self._set_param(x)

        if batch_size == -1:
            batch_size = self.N
            
        n_batch = self.N // batch_size
        if self.N % batch_size != 0: n_batch += 1

        tot_loss = 0
        self.zero_grad()        
        for k in range(n_batch):
            u_bond = self.bond_energy[k*batch_size:(k+1)*batch_size]
            u_elec = self.elec_energy[k*batch_size:(k+1)*batch_size]

            rp3angle = self._scipy_csr_to_torch_coo(self.rp3angle[k*batch_size:(k+1)*batch_size]).to(device)        
            angle = self._scipy_csr_to_torch_coo(self.angle[k*batch_size:(k+1)*batch_size]).to(device)
            dihedral = self._scipy_csr_to_torch_coo(self.dihedral[k*batch_size:(k+1)*batch_size]).to(device)
            lj = self._scipy_csr_to_torch_coo(self.lj[k*batch_size:(k+1)*batch_size]).to(device)

            u = u_bond + u_elec + rp3angle.mv(self.c3a) + angle.mv(self.ca) + \
                dihedral.mv(self.cd) + lj.mv(self.clj)

            if self.rmsd is not None:
                rmsd = self.rmsd[k*batch_size:(k+1)*batch_size].to(device)
                u = u + rmsd.mv(self.crmsd)
                
            log_p = - (u - self.F) + u.new_tensor([self.N_md/self.N_imus])
            log_q = self.log_q[k*batch_size:(k+1)*batch_size].to(device)
            target = self.target[k*batch_size:(k+1)*batch_size].to(device)

            logit = log_p - log_q
            loss = torch.nn.functional.binary_cross_entropy_with_logits(
                logit, target, reduction = 'sum')        
            loss.backward()
            tot_loss += loss.cpu().item()
            
        loss = tot_loss / self.N
        for p in self.parameters():
            if p.grad is not None:
                p.grad = p.grad / self.N

        clj = self.clj.view(self.n_lj, self.nb_lj)
        loss_wc = 0.5*weight_decay*torch.sum(clj[:,None,:].matmul(self.lj_omega).matmul(clj[:,:,None]))
        loss_wc.backward()
        
        loss += loss_wc.cpu().item()
        grad = self._gather_flat_grad()

        return loss, grad.cpu().numpy()

    def get_FF(self):
        self.bonded_parameters['reference_particle_3_angle']['U'] = torch.matmul(self.angle_grid,
                                                                                 self.c3a.cpu().data)
        nb_angle = self.angle_grid.shape[-1]                
        for i in range(len(self.bonded_parameters['angle'])):
            ca = self.ca.cpu().data[i*nb_angle:(i+1)*nb_angle]
            self.bonded_parameters['angle'][i]['U'] = torch.matmul(self.angle_grid, ca)
            
        nb_dihedral = self.dihedral_grid.shape[-1]        
        for i in range(len(self.bonded_parameters['dihedral'])):
            cd = self.cd.cpu().data[i*nb_dihedral:(i+1)*nb_dihedral]
            self.bonded_parameters['dihedral'][i]['U'] = torch.matmul(self.dihedral_grid, cd)

        clj = self.clj.cpu().data.view(self.n_lj, 1, self.nb_lj)
        ulj = torch.sum(self.lj_grid*clj, dim = -1)

        urmsd = None

        if self.rmsd is not None:
            crmsd = self.crmsd.cpu().data.view(self.n_rmsd, 1, self.nb_rmsd)
            urmsd = torch.sum(self.rmsd_grid*crmsd, dim = -1)
            
        FF = {'bonded': self.bonded_parameters,
              'lj': {'U': ulj, 'min': self.lj_min, 'max': self.lj_max},
              'rmsd': {'U': urmsd, 'min': self.rmsd_min, 'max': self.rmsd_max}}
        
        return FF

    def get_extra_state(self):
        return None
    
def gather_param(param_list, param_numel, clj_sum = False):
    assert(len(param_list) == len(param_numel))
    res = []
    clj = []
    for i in range(len(param_numel)):
        p = param_list[i]
        p_numel = param_numel[i]

        offset = 0
        for k, v in p_numel.items():
            if k != 'clj':
                res.append(p[offset:offset+v])
            else:
                clj.append(p[offset:offset+v])
            offset += v

    if clj_sum is False:
        res.append(clj[0])
    else:
        res.append(torch.sum(torch.stack(clj), dim = 0))
    return torch.cat(res, 0)

def split_param(x, param_numel):
    if type(x) == np.ndarray:
        x = torch.from_numpy(x)
    
    p_list = []
    offset = 0
    for i in range(len(param_numel)):
        p_numel = param_numel[i]
        p = []

        for k, v in p_numel.items():
            if k != 'clj':
                p.append(x[offset:offset+v])
                offset += v
            else:
                p.append(x[-v:])
        p = torch.cat(p, 0)
        p_list.append(p)
        
    return p_list

#ray.shutdown()
ray.init(ignore_reinit_error = True, _temp_dir = "/home/gridsan/dingxq/tmp/ray")    
actors = [CL_Actor.options(num_cpus = 20, num_gpus = 1).remote(name) for name in protein_names]
_ = [actor.cuda.remote() for actor in actors]
param_numel = ray.get([actor.get_named_parameter_numel.remote() for actor in actors])
param_list = ray.get([actor._gather_flat_param.remote() for actor in actors])

def compute_loss_and_grad(x, weight_decay):
    p_list = split_param(x, param_numel)    
    res = ray.get([actors[i].compute_loss_and_grad.remote(p_list[i], weight_decay/len(actors)) for i in range(len(actors))])
    loss = np.sum([l for l, g in res])
    grad_list = [torch.from_numpy(g.copy()) for l, g in res]
    grad = gather_param(grad_list, param_numel, True).numpy()    
    return loss, grad

x_init = gather_param(param_list, param_numel).cpu().numpy()
loss, grad = compute_loss_and_grad(x_init, weight_decay)
bounds = ray.get(actors[0].make_parameter_bounds.remote())
offset = 0
for k, v in param_numel[0].items():
    if k == 'clj':
        clj_bounds = bounds[offset:offset+v]
        break
    else:
        offset += v

bounds = [[None, None] for i in range(len(x_init))]
bounds[-len(clj_bounds):] = clj_bounds

x, f, d = optimize.fmin_l_bfgs_b(compute_loss_and_grad,
                                 x_init,
                                 args = [weight_decay],
                                 bounds = bounds,
                                 iprint = 1,
                                 pgtol = 1e-5,
                                 factr = 100,
                                 maxfun = 50_000,
                                 maxiter = 50_000)

FFs = ray.get([actor.get_FF.remote() for actor in actors])
state_dicts = ray.get([actor.state_dict.remote() for actor in actors])
    
param_list = ray.get([actor._gather_flat_param.remote() for actor in actors])
res = ray.get([actors[i].compute_loss_and_grad.remote(param_list[i], 0.0) for i in range(len(actors))])
loss = [l for l, g in res]
loss_total = np.sum(loss)

print("loss for each proteins", loss)
print(f"total_loss without weight decay: {loss_total:.3f}")

for i in range(len(protein_names)):
    name = protein_names[i]
    FF = FFs[i]
    state_dict = state_dicts[i]
    os.makedirs(f"./output/{name}/FF_wo_rmsd_bound_0", exist_ok = True)
    torch.save({'FF': FF, 'state_dict': state_dict, 'loss': loss[i], 'loss_total': loss_total},
               f"./output/{name}/FF_wo_rmsd_bound_0/FF_elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.pt")
    print(name)


exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=log_q_imus
#SBATCH --time=00:20:00
#SBATCH --partition=normal
#SBATCH --ntasks=1
#SBATCH --cpus-per-task=3
#SBATCH --gres=gpu:volta:1
#SBATCH --nodes=1
#SBATCH --array=0-1
#SBATCH --mem=60G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/log_q_imus_%a.txt

import numpy as np
import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import argparse
import pandas as pd
import mdtraj
from sys import exit
import os
import time
import pandas as pd
import os
import matplotlib.pyplot as plt
import pickle
from FastMBAR import *
import torch

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']

traj_imus = []
num_frames = []
stride = 10
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj = traj[::stride]
    traj_imus.append(traj)
    num_frames.append(len(traj))    
traj_imus = mdtraj.join(traj_imus)
num_frames = np.array(num_frames)

traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)

rmsd_md = mdtraj.rmsd(traj_md, traj_ref)
rmsd_imus = mdtraj.rmsd(traj_imus, traj_ref)

## make an integrator
protein_info = pd.read_csv("./info/protein_temperature_and_ionic_strength.txt",
                           index_col = 'name', comment = '#')
T = protein_info.loc[name, 'temperature']
kbT = unit.BOLTZMANN_CONSTANT_kB * T * unit.kelvin * unit.AVOGADRO_CONSTANT_NA
kbT = kbT.value_in_unit(unit.kilojoule_per_mole)

fricCoef = 1./unit.picoseconds
stepsize = 5. * unit.femtoseconds
integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)
platform = omm.Platform.getPlatformByName('Reference')
with open(f"./output/{name}/system_im.xml", 'r') as file_handle:
    xml = file_handle.read()    
system = omm.XmlSerializer.deserialize(xml)
context = omm.Context(system, integrator, platform)

energy_imus = []
for i in range(len(traj_imus)):
    context.setPositions(traj_imus.xyz[i])
    state = context.getState(getEnergy = True)
    U = state.getPotentialEnergy().value_in_unit(unit.kilojoule_per_mole)
    energy_imus.append(U)

    if (i + 1) % 10000 == 0:
        print(i, flush = True)
    
energy_imus = np.array(energy_imus)

energy_md = []
for i in range(len(traj_md)):
    context.setPositions(traj_md.xyz[i])
    state = context.getState(getEnergy = True)
    U = state.getPotentialEnergy().value_in_unit(unit.kilojoule_per_mole)
    energy_md.append(U)
energy_md = np.array(energy_md)

rmsd_centers = rmsd_centers_and_k['centers'][:, np.newaxis]
rmsd_k = rmsd_centers_and_k['k']

energy_matrix_imus = energy_imus + 0.5*rmsd_k*(rmsd_imus - rmsd_centers)**2
energy_matrix_md = energy_md + 0.5*rmsd_k*(rmsd_md - rmsd_centers)**2

energy_matrix_imus = energy_matrix_imus/kbT
energy_matrix_md = energy_matrix_md/kbT

fastmbar = FastMBAR(energy_matrix_imus, num_frames, verbose = True, cuda = True)
log_q_imus = fastmbar.log_prob_mix

biased_energy = energy_matrix_md + fastmbar.bias_energy.reshape((-1,1))
biased_energy_min = np.min(biased_energy, 0, keepdims = True)        
log_q_md = np.log(np.sum(np.exp(-(biased_energy - biased_energy_min)), 0)) - biased_energy_min.reshape(-1)

log_q_imus = torch.from_numpy(log_q_imus)
log_q_md = torch.from_numpy(log_q_md)

torch.save({'md': log_q_md, 'imus': log_q_imus},
           f"./output/{name}/log_q_imus.pt")



import mdtraj
import argparse
import math
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
import itertools
import pandas as pd

weight_decay_list = [0.0,
                     1e-10, 5e-10,  
                     1e-9 , 5e-9 ,
                     1e-8 , 5e-8 ,
                     1e-7 , 5e-7 ,
                     1e-6 , 5e-6 ,
                     1e-5 , 5e-5 ,
                     1e-4 , 5e-4 ,
                     1e-3 , 5e-3 ,
                     1e-2 , 5e-2 ,
                     1e-1 ]

## read system
name = "2JOF"
# full_include_rmsd = True
# full_weight_decay = 2e-7

full_include_rmsd = False
full_weight_decay = 4e-7


psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)    
traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
rmsd_md = mdtraj.rmsd(traj_md, traj_ref)

pdf_hist =  PdfPages(f"./output/{name}/rmsd_full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay:.3E}_nnforce_hist.pdf")
pdf_traj =  PdfPages(f"./output/{name}/rmsd_full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay:.3E}_nnforce_traj.pdf")

for weight_decay in weight_decay_list:

    print(f"name: {name}, weight_decay: {weight_decay}", flush = True)

    traj_cg_NVT = mdtraj.load_dcd(f"./output/{name}/nnforce_NVT/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_weight_decay_{weight_decay:.3E}.dcd", psf)
    rmsd_cg_NVT = mdtraj.rmsd(traj_cg_NVT, traj_ref)        

    fig = plt.figure()
    fig.clf()

    plt.hist(rmsd_md, bins = 40, alpha = 0.5, label = 'md', density = True)
    plt.hist(rmsd_cg_NVT, bins = 40, alpha = 0.5, label = 'cg_NVT', density = True)

    plt.legend()
    plt.title(f"weight_decay: {weight_decay:.3E}")
    pdf_hist.savefig()
    plt.close()

    fig = plt.figure(figsize = (6.4*4, 4.8))
    fig.clf()

    plt.plot(rmsd_md[::len(rmsd_md)//10000], alpha = 0.5, label = 'md')
    plt.plot(rmsd_cg_NVT[::len(rmsd_cg_NVT)//10000], alpha = 0.5, label = 'cg_NVT')

    plt.legend()
    plt.title(f"weight_decay: {weight_decay:.3E}")
    pdf_traj.savefig()
    plt.close()

pdf_hist.close()
pdf_traj.close()

exit()



import argparse
import numpy as np
import torch
torch.set_default_dtype(torch.float64)
import pickle
from scipy import optimize
from sys import exit
import mdtraj
import pandas as pd
import simtk.unit as unit
from FastMBAR import *
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
sys.path.append("./script")
from NNForce import *
import os
from functools import reduce
import matplotlib.pyplot as plt

name = '2JOF'
full_include_rmsd = False
full_weight_decay = 4e-7

print(f"full rmsd: {full_include_rmsd}, weight_decay: {full_weight_decay:.3E}")

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
hidden_size = 32
n = psf.n_atoms
index = torch.LongTensor([[i,j] for i in range(psf.n_atoms) for j in range(i+4, psf.n_atoms)])
nnforce_net_1 = NNForceNet(n - 2,
                           n - 3,
                           len(index),
                           hidden_size)

nnforce_net_2 = NNForceNet(n - 2,
                           n - 3,
                           len(index),
                           hidden_size)

weight_decay = 1e-3
data = torch.load(f"./output/{name}/FF/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_NNForce_weight_decay_{weight_decay:.3E}_different_initial_parameters.pth")
state_dict = data['state_dict']
nnforce_net_1.load_state_dict(state_dict)

data = torch.load(f"./output/{name}/FF/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_NNForce_weight_decay_{weight_decay:.3E}.pth")
state_dict = data['state_dict']
nnforce_net_2.load_state_dict(state_dict)


p1 = torch.cat([p.data.flatten() for p in nnforce_net_1.parameters()])
p2 = torch.cat([p.data.flatten() for p in nnforce_net_2.parameters()])

fig = plt.figure()
fig.clf()
plt.plot(p1.numpy(), p2.numpy(), '.', alpha = 0.5)
plt.savefig('./output/plots/compare_nnforce_parameters_learned_with_different_initial_parameters.png')



import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'x-large'
mpl.rcParams['xtick.labelsize'] = 'x-large'
mpl.rcParams['ytick.labelsize'] = 'x-large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm
import pickle
from matplotlib.backends.backend_pdf import PdfPages

name = "2JOF"
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
resnames = [residue.name for residue in psf.residues]
for i in range(len(resnames)):
    if resnames[i] == "NLE":
        resnames[i] = "ILE"

## load spline coefficients
data = torch.load(f"./output/{name}/FF/FF_rmsd_True_weight_decay_{2e-7:.3E}.pt")
U = data['FF']['lj']['U'].numpy()
r_min, r_max = data['FF']['lj']['min'], data['FF']['lj']['max']
r = np.linspace(r_min, r_max, U.shape[1])
pdf = PdfPages(f"./output/plots/{name}_lj_all.pdf")
for i in range(U.shape[0]):
    fig = plt.figure()
    plt.plot(r, U[i])
    plt.ylim(-3, 5)
    plt.title(f"index_{i}")
    pdf.savefig()
    print(i)
pdf.close()    

pdf = PdfPages(f"./output/plots/{name}_lj_selected.pdf")
fig = plt.figure()
for i in [103, 123, 128]:    
    plt.plot(r, U[i], linewidth = 4, alpha = 0.6)
plt.ylim(-3, 5)
pdf.savefig()

fig = plt.figure()
plt.plot(r, U[108], linewidth = 4)
plt.ylim(-3, 5)
pdf.savefig()

pdf.close()
exit()


nrow = 6
ncol = 4

fig = plt.figure(np.random.randint(0, 1000), figsize = (6.4*ncol, 4.8*nrow))
fig.clf()
for k in range(len(pair_indices)):
    plt.subplot(nrow, ncol, k+1)
    pair = pair_indices[k]
    i,j = pair_indices[k]

    aa1, aa2 = resnames[i], resnames[j]
    if resnames[j] < resnames[i]:
        aa1, aa2 = resnames[j], resnames[i]
    
    r_min, r_max = LJ_parameters[pair]['r_min'], LJ_parameters[pair]['r_max']
    ulj = LJ_parameters[pair]['ulj']
    r = np.linspace(r_min , r_max, len(ulj))
    plt.plot(r, ulj, linewidth = 6.0)
    plt.xlim(0, r_max)
    plt.ylim(ulj.min()-1, 7)

    plt.title(f"{i}-{j}:{aa1}-{aa2}")
    print(k)

plt.tight_layout()    
fig.savefig(f"./output/plots/{name}_lj.eps")

exit()



import numpy as np
import openmm as omm
import openmm.unit as unit
import openmm.app as ommapp
import math
import mdtraj
import pickle
from sys import exit
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
import torch

def make_system(masses,
                coor_transformer,
                bonded_parameters,
                T,
                exclude_angle_particle_index = [],
                exclude_dihedral_particle_index = [],
                jacobian = False):
    
    ## add particles
    system = omm.System()
    for m in masses:
        system.addParticle(m)

    Kb = unit.BOLTZMANN_CONSTANT_kB*unit.AVOGADRO_CONSTANT_NA
    Kb = Kb.value_in_unit(unit.kilojoule_per_mole/unit.kelvin)

    ## bond force between reference particle 2 and reference particle 1
    bond_force_ref_2 = omm.CustomBondForce("0.5*kb*(r - b0)^2*Kb*T")
    bond_force_ref_2.addGlobalParameter("Kb", Kb)
    bond_force_ref_2.addGlobalParameter("T", T)
    bond_force_ref_2.addPerBondParameter('b0')
    bond_force_ref_2.addPerBondParameter('kb')
    bond_force_ref_2.addBond(
        coor_transformer.ref_particle_1,
        coor_transformer.ref_particle_2,
        [bonded_parameters['reference_particle_2_bond']['b0'],
         bonded_parameters['reference_particle_2_bond']['kb']]
    )
    
    bond_force_ref_2.setForceGroup(0)
    system.addForce(bond_force_ref_2)

    ## bond force between reference particle 3 and reference particle 1
    if jacobian:
        bond_force_ref_3 = omm.CustomBondForce("(0.5*kb*(r - b0)^2 + log(r))*Kb*T")
    else:
        bond_force_ref_3 = omm.CustomBondForce("(0.5*kb*(r - b0)^2)*Kb*T")        
    bond_force_ref_3.addGlobalParameter("Kb", Kb)
    bond_force_ref_3.addGlobalParameter("T", T)
    bond_force_ref_3.addPerBondParameter('b0')
    bond_force_ref_3.addPerBondParameter('kb')
    bond_force_ref_3.addBond(
        coor_transformer.ref_particle_1,
        coor_transformer.ref_particle_3,
        [bonded_parameters['reference_particle_3_bond']['b0'],
         bonded_parameters['reference_particle_3_bond']['kb']]
    )
    bond_force_ref_3.setForceGroup(0)
    system.addForce(bond_force_ref_3)

    ## bond force between other particles
    if jacobian:
        bond_force = omm.CustomBondForce("(0.5*kb*(r - b0)^2 + 2*log(r))*Kb*T")
    else:
        bond_force = omm.CustomBondForce("(0.5*kb*(r - b0)^2)*Kb*T")        
    bond_force.addGlobalParameter("Kb", Kb)
    bond_force.addGlobalParameter("T", T)
    bond_force.addPerBondParameter('b0')
    bond_force.addPerBondParameter('kb')
    for i in range(len(coor_transformer.particle_visited_in_order)):
        p = coor_transformer.particle_visited_in_order[i]
        p1 = coor_transformer.bond_particle_idx[p][0]
        p2 = coor_transformer.bond_particle_idx[p][1]
        bond_force.addBond(
            p1,
            p2,
            [float(bonded_parameters['bond']['b0'][i].item()),
             float(bonded_parameters['bond']['kb'][i].item())]
        )
    bond_force.setForceGroup(0)
    system.addForce(bond_force)
    
    ## custom angle force
    if 'a0' in bonded_parameters['reference_particle_3_angle'] and 'ka' in bonded_parameters['reference_particle_3_angle']:
        angle_force_between_reference_particles = omm.CustomAngleForce("0.5*ka*(theta - a0)^2*Kb*T")
        angle_force_between_reference_particles.addGlobalParameter('Kb', Kb)
        angle_force_between_reference_particles.addGlobalParameter('T', T)
        angle_force_between_reference_particles.addPerAngleParameter('a0')
        angle_force_between_reference_particles.addPerAngleParameter('ka')
        if not (coor_transformer.ref_particle_1 in exclude_angle_particle_index and
                coor_transformer.ref_particle_2 in exclude_angle_particle_index and
                coor_transformer.ref_particle_3 in exclude_angle_particle_index):
            angle_force_between_reference_particles.addAngle(
                coor_transformer.ref_particle_2,
                coor_transformer.ref_particle_1,
                coor_transformer.ref_particle_3,
                [bonded_parameters['reference_particle_3_angle']['a0'],
                 bonded_parameters['reference_particle_3_angle']['ka']]
            )
        angle_force_between_reference_particles.setForceGroup(0)
        system.addForce(angle_force_between_reference_particles)

        if jacobian:
            angle_force = omm.CustomAngleForce("(alpha*0.5*ka*(theta - a0)^2 + log(sin(pi - theta)))*Kb*T; pi=3.14159265357")
        else:
            angle_force = omm.CustomAngleForce("(alpha*0.5*ka*(theta - a0)^2)*Kb*T")
        angle_force.addGlobalParameter('Kb', Kb)
        angle_force.addGlobalParameter('T', T)
        angle_force.addPerAngleParameter('a0')
        angle_force.addPerAngleParameter('ka')
        angle_force.addPerAngleParameter('alpha')

        for pi in range(len(coor_transformer.particle_visited_in_order)):
            p = coor_transformer.particle_visited_in_order[i]
            p1 = coor_transformer.angle_particle_idx[p][0]
            p2 = coor_transformer.angle_particle_idx[p][1]
            p3 = coor_transformer.angle_particle_idx[p][2]
            if (p1 in exclude_angle_particle_index and
                p2 in exclude_angle_particle_index and
                p3 in exclude_angle_particle_index):
                alpha = 0.0
            else:
                alpha = 1.0
                
            angle_force.addAngle(
                p1,
                p2,
                p3,
                [float(bonded_parameters['angle']['a0'][i].item()),
                 float(bonded_parameters['angle']['ka'][i].item()),
                 alpha]
            )
        angle_force.setForceGroup(0)        
        system.addForce(angle_force)
        
    elif 'U' in bonded_parameters['reference_particle_3_angle']:
        ## angle force between reference particles
        f = omm.Continuous1DFunction(bonded_parameters['reference_particle_3_angle']['U'].tolist(), 0.0, np.pi, periodic = False)
        angle_force_between_reference_particles = omm.CustomCompoundBondForce(3, f"ua_between_ref(angle(p1, p2, p3))*Kb*T")
        angle_force_between_reference_particles.addGlobalParameter('Kb', Kb)
        angle_force_between_reference_particles.addGlobalParameter('T', T)
        angle_force_between_reference_particles.addTabulatedFunction(f"ua_between_ref", f)
        if not (coor_transformer.ref_particle_1 in exclude_angle_particle_index and
                coor_transformer.ref_particle_2 in exclude_angle_particle_index and
                coor_transformer.ref_particle_3 in exclude_angle_particle_index):            
            angle_force_between_reference_particles.addBond(
                [coor_transformer.ref_particle_2,
                 coor_transformer.ref_particle_1,
                 coor_transformer.ref_particle_3])
        angle_force_between_reference_particles.setForceGroup(0)
        system.addForce(angle_force_between_reference_particles)

        ## angle force between other particles        
        angle_parameters = bonded_parameters['angle']
        ua = torch.stack([angle_parameters[i]['U'] for i in range(len(angle_parameters))]).numpy()
        func = omm.Continuous2DFunction(xsize = ua.shape[0],
                                        ysize = ua.shape[1],
                                        values = ua.reshape(-1, order = 'F'),
                                        xmin = 0.0, xmax = float(ua.shape[0] - 1),
                                        ymin = 0.0, ymax = np.pi,
                                        periodic = False)
        if jacobian:
            angle_force = omm.CustomCompoundBondForce(
                3,
                f"(alpha*ua(idx, angle(p1, p2, p3)) + log(sin(pi - angle(p1, p2, p3))))*Kb*T; pi = 3.14159265357"
            )
        else:
            angle_force = omm.CustomCompoundBondForce(
                3,
                f"(alpha*ua(idx, angle(p1, p2, p3)))*Kb*T"
            )
            
        angle_force.addGlobalParameter('Kb', Kb)
        angle_force.addGlobalParameter('T', T)
        angle_force.addTabulatedFunction("ua", func)
        angle_force.addPerBondParameter('idx')
        angle_force.addPerBondParameter('alpha')
        
        for i in range(len(coor_transformer.particle_visited_in_order)):
            p = coor_transformer.particle_visited_in_order[i]
            p1, p2, p3 = coor_transformer.angle_particle_idx[p]
            if (p1 in exclude_angle_particle_index and
                p2 in exclude_angle_particle_index and
                p3 in exclude_angle_particle_index):
                alpha = 0.0
            else:
                alpha = 1.0
            angle_force.addBond([p1, p2, p3], [float(i), alpha])            
        angle_force.setForceGroup(0)
        system.addForce(angle_force)
        
    ## torsion force
    dihedral_parameters = bonded_parameters['dihedral']
    ud = torch.stack([dihedral_parameters[i]['U'] for i in range(len(dihedral_parameters))] +
                     [dihedral_parameters[0]['U']]).numpy()
    func = omm.Continuous2DFunction(xsize = ud.shape[0],
                                    ysize = ud.shape[1],
                                    values = ud.reshape(-1, order = 'F'),
                                    xmin = 0.0, xmax = float(ud.shape[0] - 1),
                                    ymin = -np.pi, ymax = np.pi,
                                    periodic = True)
    dihedral_force = omm.CustomCompoundBondForce(4, f"(alpha*ud(idx, dihedral(p1, p2, p3, p4)))*Kb*T")
    dihedral_force.addGlobalParameter('Kb', Kb)
    dihedral_force.addGlobalParameter('T', T)
    dihedral_force.addTabulatedFunction("ud", func)
    dihedral_force.addPerBondParameter('idx')
    dihedral_force.addPerBondParameter('alpha')    

    for i in range(len(coor_transformer.particle_visited_in_order)):
        p = coor_transformer.particle_visited_in_order[i]
        p1, p2, p3, p4 = coor_transformer.dihedral_particle_idx[p]
        if (p1 in exclude_dihedral_particle_index and
            p2 in exclude_dihedral_particle_index and
            p3 in exclude_dihedral_particle_index and                
            p4 in exclude_dihedral_particle_index):
            alpha = 0.0
        else:
            alpha = 1.0
        dihedral_force.addBond([p1, p2, p3, p4], [float(i), alpha])
    dihedral_force.setForceGroup(0)
    system.addForce(dihedral_force)

    system.addForce(omm.CMMotionRemover())
    
    return system



import pickle
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
mpl.rc('font', size = 16)
mpl.rc('axes', titlesize = 'large', labelsize = 'large')
mpl.rc('xtick', labelsize = 'large')
mpl.rc('ytick', labelsize = 'large')
from matplotlib.backends.backend_pdf import PdfPages
import os
import numpy as np
import pandas as pd
from sys import exit
from matplotlib.pyplot import cm
import argparse
import sys
import torch
from functions import *
import scipy.cluster as cluster

with open(f"./output/common/LJ_rmin.pkl", 'rb') as file_handle:
    r_min_dict = pickle.load(file_handle)
aa_pairs = list(r_min_dict.keys())
aa_pairs.sort()

AA_info = pd.read_csv("./info/amino_acids_with_learned_sigmas.csv",
                      index_col = 'name')
AA_names = list(AA_info.index)
AA_names.pop(AA_names.index('CYS'))
# AA_name_dict = {'G': 'GLY', 'P': 'PRO', 'A': 'ALA', 'V': 'VAL',
#                 'L': 'LEU', 'I': 'ILE', 'M': 'MET', 'C': 'CYS',
#                 'F': 'PHE', 'Y': 'TYR', 'W': 'TRP', 'H': 'HIS',
#                 'K': 'LYS', 'R': 'ARG', 'Q': 'GLN', 'N': 'ASN',
#                 'E': 'GLU', 'D': 'ASP', 'S': 'SER', 'T': 'THR'}

AA_name_dict = {'G': 'GLY', 'P': 'PRO', 'A': 'ALA', 'V': 'VAL',
                'L': 'LEU', 'I': 'ILE', 'M': 'MET',
                'F': 'PHE', 'Y': 'TYR', 'W': 'TRP', 'H': 'HIS',
                'K': 'LYS', 'R': 'ARG', 'Q': 'GLN', 'N': 'ASN',
                'E': 'GLU', 'D': 'ASP', 'S': 'SER', 'T': 'THR'}
NUM_COLORS = 20

name = "CLN025"
weight_decay = 1e-6

elec_type = 'DH_2'
ss_type = 'simple'

data = torch.load(f"./output/{name}/FF/FF_elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.pt",
                  map_location=torch.device('cpu'))

# data = torch.load(f"./output/{name}/FF_bound_0/FF_elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.pt",
#                   map_location=torch.device('cpu'))

FF = data['FF']
bonded_ff, lj_ff, rmsd_ff = FF['bonded'], FF['lj'], FF['rmsd']

ncols = 4
nrows = 5
fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
for aa1 in AA_names:
    print(aa1)
    ax = plt.subplot(nrows, ncols, AA_names.index(aa1) + 1)
    colors = iter([cm.tab20(i) for i in range(20)])

    lines = []
    for aa2 in AA_names:
        aa = [aa1, aa2]
        aa.sort()
        if (aa[0], aa[1]) in aa_pairs:
            idx = aa_pairs.index((aa[0], aa[1]))
            r = np.linspace(lj_ff['min'], lj_ff['max'], lj_ff['U'].shape[-1])
            line = ax.plot(r, lj_ff['U'][idx],
                           linewidth = 2.0,
                           label = aa2,
                           color = next(colors))
            lines.append(line)
    plt.ylim(-4.5, 4.5)
    plt.xlim(0, lj_ff['max'])
    #plt.title(aa1)
    plt.text(0.1, 3, aa1)
    #plt.xlabel('distance (nm)')
    
ax = plt.subplot(nrows, ncols, 20)
colors = iter([cm.tab20(i) for i in range(20)])
#ax.legend(lines, AA_names)
for aa2 in AA_names:
    plt.plot([], [],
            linewidth = 2.0,
            label = aa2,
            color = next(colors))
ax.legend(ncol = 3)
ax.axis('off')

#plt.tight_layout()
plt.savefig("./output/figures/lj_potential.pdf")    

# pair_names = [['TRP', 'TYR'],
#               ['THR', 'TYR'],
#               ['MET', 'SER'],
#               ['PRO', 'TRP'],
#               ['PRO', 'THR']]

pair_names = [['ALA', 'LYS'],
              ['ILE', 'VAL'],
              ['TRP', 'TYR'],
              ['ASP', 'THR']]

fig = plt.figure()
fig.clf()
for name in pair_names:
    name.sort()
    idx = aa_pairs.index(tuple(name))
    r = np.linspace(lj_ff['min'], lj_ff['max'], lj_ff['U'].shape[-1])
    plt.plot(r, lj_ff['U'][idx], linewidth = 2.0, label = '-'.join(name))
    plt.ylim(-4.5, 4.5)
plt.legend(ncol = 2)
plt.xlabel('Distance (nm)')
plt.ylabel('Energy (kJ/mol)')
plt.tight_layout()
plt.savefig("./output/figures/lj_potential_selected.eps")    

exit()


# pdf = PdfPages("./output/figures/lj_potential_seperate.pdf")
# for aa1 in AA_names:
#     print(aa1)
#     ncols = 5
#     nrows = 4

#     fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
#     fig.suptitle(aa1)
    
#     for aa2 in AA_names:
#         plt.subplot(nrows, ncols, AA_names.index(aa2) + 1)
#         colors = iter([cm.tab20(i) for i in range(20)])
#         aa = [aa1, aa2]
#         aa.sort()
#         print(aa)
#         if (aa[0], aa[1]) in aa_pairs:
#             idx = aa_pairs.index((aa[0], aa[1]))
#             r = np.linspace(lj_ff['min'], lj_ff['max'], lj_ff['U'].shape[-1])
#             plt.plot(r, lj_ff['U'][idx],
#                     linewidth = 3.0,
#                     label = aa2,
#                     color = next(colors))
#             plt.legend()
            
#         plt.ylim(-4.5, 4.5)
#         plt.xlim(0, lj_ff['max'])
#     pdf.savefig()
# pdf.close()

aa_names = 'RHKDESTNQGPAVILMFYW'
AA_names = [AA_name_dict[n] for n in aa_names]
        
E_min = np.zeros((len(AA_names), len(AA_names)))
for i in range(len(AA_names)):
    for j in range(len(AA_names)):
        aa1, aa2 = AA_names[i], AA_names[j]
        aa = [aa1, aa2]
        aa.sort()
        if (aa[0], aa[1]) in aa_pairs:
            idx = aa_pairs.index((aa[0], aa[1]))
            E_min[i,j] = lj_ff['U'][idx].min().item()
        elif (aa[0], aa[1]) == ('TRP', 'TRP'):
            idx = aa_pairs.index(('TRP', 'TYR'))
            E_min[i,j] = lj_ff['U'][idx].min().item()
        else:
            E_min[i,j] = None

fig = plt.figure(figsize = (6.4*2, 4.8*2))
fig.clf()
ax = plt.subplot()
im, cbar = heatmap(E_min, AA_names, AA_names, ax = ax, origin = 'lower')
fig.tight_layout()
plt.savefig(f"./output/figures/lj_heatmap.pdf")

exit()


linkage = cluster.hierarchy.linkage(E_min)


exit()

AA_names = [AA_name_dict[n] for n in 'MRGDPNKSAFVLTYHEQIWC']
E_min = np.zeros((len(AA_names), len(AA_names)))
for i in range(len(AA_names)):
    for j in range(len(AA_names)):
        aa1, aa2 = AA_names[i], AA_names[j]
        aa = [aa1, aa2]
        aa.sort()
        if (aa[0], aa[1]) in aa_pairs:
            idx = aa_pairs.index((aa[0], aa[1]))
            E_min[i,j] = lj_ff['U'][idx].min().item()
        else:
            E_min[i,j] = None

fig = plt.figure(figsize = (6.4*2, 4.8*2))
fig.clf()
ax = plt.subplot()
im, cbar = heatmap(E_min, AA_names, AA_names, ax = ax, origin = 'lower')
fig.tight_layout()
plt.savefig(f"./output/figures/lj_heatmap_MOFF_order.pdf")

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=make_system
#SBATCH --time=01:00:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-3
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/make_system_%a.txt

import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import pickle
import math
from sys import exit
from collections import defaultdict
import numpy as np
import mdtraj
from collections import defaultdict
import pandas as pd
import argparse
from scipy.interpolate import BSpline
import os
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG/")
from CLCG.utils.make_system import *
import itertools

protein_names = pd.read_csv("./info/protein_names.txt",
                            comment = "#",
                            header = None)
protein_names = protein_names.iloc[0:2, 0].tolist()
flag_rmsd = [False, True]
# weight_decay_list = [1e-7, 5e-7,
#                      1e-6, 5e-6,
#                      1e-5, 5e-5,
#                      1e-4, 5e-4,
#                      1e-3, 5e-3]
weight_decay_list = [2e-7, 3e-7, 4e-7]
options = list(itertools.product(protein_names, flag_rmsd, weight_decay_list))

for rank in range(len(options)):    
    name, include_rmsd, weight_decay = options[rank]
    print(f"name: {name}, rmsd: {include_rmsd}, weight_decay: {weight_decay}", flush = True)

    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    resnames = [residue.name for residue in psf.residues]
    for i in range(len(resnames)):
        if resnames[i] == "NLE":
            resnames[i] = "ILE"
    nres = len(resnames)

    protein_info = pd.read_csv("./info/protein_temperature_and_ionic_strength.txt",
                               index_col = 'name',
                               comment = '#')
    T = protein_info.loc[name, 'temperature']
    ionic_strength = protein_info.loc[name, 'ionic_strength']
    
    kbT = unit.BOLTZMANN_CONSTANT_kB * T *unit.kelvin * unit.AVOGADRO_CONSTANT_NA
    kbT_kJ_per_mole = kbT.value_in_unit(unit.kilojoule_per_mole)
    beta = 1/kbT_kJ_per_mole

    Kb = unit.BOLTZMANN_CONSTANT_kB*unit.AVOGADRO_CONSTANT_NA
    Kb = Kb.value_in_unit(unit.kilojoule_per_mole/unit.kelvin)

    with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
        coor_transformer = pickle.load(file_handle)

    AA_info = pd.read_csv("./info/amino_acids_with_learned_sigmas.csv", index_col = 'name')
    masses = [AA_info.loc[r.name, 'mass'] for r in psf.residues]

    data = torch.load(f"./output/{name}/FF/FF_rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.pt")
    FF = data['FF']
    bonded_ff, lj_ff, rmsd_ff = FF['bonded'], FF['lj'], FF['rmsd']

    ## bonded terms
    system = make_system(masses,
                         coor_transformer,
                         bonded_ff,
                         protein_info.loc[name, 'temperature'],
                         jacobian = False)

    ## electrostatic
    def make_custom_elec_force(T, ionic_strength):
        formula = ["ONE_4PI_EPS0*charge1*charge2/(r*eps)*exp(-r/lambda_D)",
                   "lambda_D = sqrt(eps*constant*T/ionic_strength)",
                   "constant = 2.3811197070710097 / 10^6",
                   "eps = 78.4",
                   "ONE_4PI_EPS0 = 138.935456",
        ]

        custom_elec_force = omm.CustomNonbondedForce(";".join(formula))
        custom_elec_force.addGlobalParameter("T", T)
        custom_elec_force.addGlobalParameter("ionic_strength", ionic_strength)
        custom_elec_force.addPerParticleParameter('charge')
        return custom_elec_force

    custom_elec_force = make_custom_elec_force(T, ionic_strength)
    custom_elec_force.setForceGroup(1)
    for i in range(psf.n_residues):
        resname = psf.residue(i).name
        charge = AA_info.loc[resname, 'charge']
        custom_elec_force.addParticle([charge])
    bonds = [[i, i+1] for i in range(psf.n_residues - 1)]
    custom_elec_force.createExclusionsFromBonds(bonds, 3)
    system.addForce(custom_elec_force)
    
    ## LJ
    aa_pairs = [(i,j) for i in range(psf.n_residues)
                      for j in range(i+4, psf.n_residues)]    
    aa_pairs.sort()

    ulj = lj_ff['U'].numpy()
    lj_min = lj_ff['min']
    lj_max = lj_ff['max']    
    func = omm.Continuous2DFunction(xsize = ulj.shape[0],
                                    ysize = ulj.shape[1],
                                    values = ulj.reshape(-1, order = 'F'),
                                    xmin = 0., xmax = float(ulj.shape[0] - 1),
                                    ymin = lj_min, ymax = lj_max)

    LJ_force = omm.CustomCompoundBondForce(2, f"Kb*T*U_LJ(idx, r); r = distance(p1,p2)")
    LJ_force.addGlobalParameter('Kb', Kb)
    LJ_force.addGlobalParameter('T', T)
    LJ_force.addTabulatedFunction(f"U_LJ", func)
    LJ_force.addPerBondParameter('idx')

    res_names = [res.name for res in psf.residues]    
    for i in range(len(res_names)):
        for j in range(i+4, len(res_names)):
            idx = float(aa_pairs.index((i,j)))
            LJ_force.addBond([i,j], [idx])
    LJ_force.setForceGroup(2)        
    system.addForce(LJ_force)

    ## rmsd
    if include_rmsd:
        ss = pd.read_table("./info/secondary_structure.csv",
                           index_col = 'name', header = 0)
        ss = eval(ss.loc[name, 'ss'])
        traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

        rmsd_min = rmsd_ff['min']
        rmsd_max = rmsd_ff['max']
        rmsd_U = rmsd_ff['U']

        for idx_ss in range(len(ss)):
            particle_index = [list(range(start-1, end)) for start, end in ss[idx_ss]]
            particle_index = list(itertools.chain(*particle_index))

            rmsd_force = omm.RMSDForce(traj_ref.xyz[0], particle_index)
            cv_force = omm.CustomCVForce("Kb*T*u_rmsd(rmsd)")
            cv_force.addCollectiveVariable('rmsd', rmsd_force)
            u_rmsd = omm.Continuous1DFunction(rmsd_U[idx_ss].numpy(),
                                              rmsd_min[idx_ss].item(),
                                              rmsd_max[idx_ss].item())
            cv_force.addTabulatedFunction('u_rmsd', u_rmsd)
            cv_force.addGlobalParameter('Kb', Kb)
            cv_force.addGlobalParameter('T', T)
            system.addForce(cv_force)

    xml = omm.XmlSerializer.serialize(system)
    os.makedirs(f"./output/{name}/full_system", exist_ok = True)
    with open(f"./output/{name}/full_system/rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.xml", 'w') as file_handle:
        file_handle.write(xml)

    fricCoef = 1./unit.picoseconds
    stepsize = 2. * unit.femtoseconds
    integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)

    platform = omm.Platform.getPlatformByName('Reference')
    context = omm.Context(system, integrator, platform)
    traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)
    context.setPositions(traj_ref.xyz[0])
    state = context.getState(getEnergy = True)
    potential_energy = state.getPotentialEnergy()
    print(potential_energy)

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=tica
#SBATCH --time=3:00:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --array=0-19
#SBATCH --mem=100G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/tica_%a.txt

import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'x-large'
mpl.rcParams['xtick.labelsize'] = 'x-large'
mpl.rcParams['ytick.labelsize'] = 'x-large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm
import math
from sklearn.mixture import GaussianMixture
import ray
import copy

name = "2JOF"

weight_decay_list = [0.0,
                     1e-10, 5e-10,  
                     1e-9 , 5e-9 ,
                     1e-8 , 5e-8 ,
                     1e-7 , 5e-7 ,
                     1e-6 , 5e-6 ,
                     1e-5 , 5e-5 ,
                     1e-4 , 5e-4 ,
                     1e-3 , 5e-3 ,
                     1e-2 , 5e-2 ,
                     1e-1 ]

#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
#weight_decay = weight_decay_list[job_idx]
weight_decay = 1e-3

# full_include_rmsd = False
# full_weight_decay = 4e-7

full_include_rmsd = False
full_weight_decay = 4e-7

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
traj_reference = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_md = mdtraj.compute_distances(traj_md, index)

tica = pyemma.coordinates.tica(distances_md, lag = 1000, dim = 2, kinetic_map = False)
x_md = tica.get_output()[0]

x = np.copy(x_md[x_md[:,1] < 5])
gm = GaussianMixture(n_components = 2)
gm.fit(x)

x = np.copy(x_md[(x_md[:,1] > 6) & (x_md[:,0] > 0)])
centers = np.concatenate([gm.means_, x.mean(0, keepdims = True)])
print(centers)

traj_md = traj_md.superpose(traj_reference, 0)
for i in range(centers.shape[0]):
    d = np.sqrt(np.sum((x_md - centers[i])**2, -1))
    idx = d.argsort()
    idx = idx[0:1]
    traj_md[idx].save_dcd(f"./output/plots/{name}_center_{i}.dcd")

exit()

flag_0 = x_md[:,0] < -1
flag_1 = (x_md[:,0] >= -1) & (x_md[:,1] < 5)
flag_2 = (x_md[:,0] >= -1) & (x_md[:,1] >= 5)

traj_md_0 = traj_md[flag_0]
traj_md_1 = traj_md[flag_1]
traj_md_2 = traj_md[flag_2]

traj_md_0_ref = ray.put(traj_md_0)
traj_md_1_ref = ray.put(traj_md_1)
traj_md_2_ref = ray.put(traj_md_2)

@ray.remote
def compute_rmsd(traj, i, j):
    traj = copy.deepcopy(traj)
    return [np.mean(mdtraj.rmsd(traj, traj, k)) for k in range(i,j) if k < traj.n_frames]

stride = 1000
rmsd_0 = ray.get([compute_rmsd.remote(traj_md_0_ref, i*stride, (i+1)*stride) for i in range(traj_md_0.n_frames//stride + 1)])
rmsd_1 = ray.get([compute_rmsd.remote(traj_md_1_ref, i*stride, (i+1)*stride) for i in range(traj_md_1.n_frames//stride + 1)])
rmsd_2 = ray.get([compute_rmsd.remote(traj_md_2_ref, i*stride, (i+1)*stride) for i in range(traj_md_2.n_frames//stride + 1)])

rmsd_0 = list(itertools.chain(*rmsd_0))
rmsd_1 = list(itertools.chain(*rmsd_1))
rmsd_2 = list(itertools.chain(*rmsd_2))

traj_md_0[np.argmin(rmsd_0)].save_xyz(f"./output/plots/{name}_center_0.xyz")
traj_md_1[np.argmin(rmsd_1)].save_xyz(f"./output/plots/{name}_center_1.xyz")
traj_md_2[np.argmin(rmsd_2)].save_xyz(f"./output/plots/{name}_center_2.xyz")




#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=rmsd_imus
#SBATCH --time=00:20:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --cpus-per-task=1
#SBATCH --nodes=1
#SBATCH --array=0-1
#SBATCH --mem=60G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/rmsd_imus_%a.txt

import numpy as np
import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import argparse
import pandas as pd
import mdtraj
from sys import exit
import os
import time
import pandas as pd
import os
import matplotlib.pyplot as plt
import pickle

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
job_idx = 1
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
    
size = rmsd_centers_and_k['size']
rmsd = {}
for rank in range(size):
    print(f"rank: {rank}")
    traj_imus = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    rmsd[rank] = mdtraj.rmsd(traj_imus, traj_ref)

fig = plt.figure(figsize = (6.4*4, 4.8))
fig.clf()
for rank in range(size):
    plt.hist(rmsd[rank], 30, density = True, alpha = 0.4)
    #plt.axvline(rmsd_centers_and_k['centers'][rank], linestyle = '--')
plt.xlabel('RMSD (nm)')
plt.ylabel('probability density')
plt.savefig(f"./output/{name}/rmsd_hist_imus.pdf")



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=make_system
#SBATCH --time=01:00:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-3
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/make_system_%a.txt

import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import pickle
import math
from sys import exit
from collections import defaultdict
import numpy as np
import mdtraj
from collections import defaultdict
import pandas as pd
import argparse
import os
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
import itertools
import torch
torch.set_default_dtype(torch.double)
sys.path.append("./script")
from NNForce import *
from openmmtorch import TorchForce


name = '2JOF'
weight_decay_list = [1e-3]
# full_include_rmsd = True
# full_weight_decay = 2e-7

full_include_rmsd = False
full_weight_decay = 4e-7

for weight_decay in weight_decay_list:

    ## read system
    with open(f"./output/{name}/full_system/rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay:.3E}.xml", 'r') as file_handle:
        xml = file_handle.read()    
    system = omm.XmlSerializer.deserialize(xml)

    protein_info = pd.read_csv(
        "./info/protein_temperature_and_ionic_strength.txt",
        index_col = 'name',
        comment = '#')
    T = protein_info.loc[name, 'temperature']

    with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
        coor_transformer = pickle.load(file_handle)

    data = torch.load(f"./output/{name}/FF/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_NNForce_weight_decay_{weight_decay:.3E}_different_initial_parameters.pth")
    hidden_size = data['hidden_size']
    state_dict = data['state_dict']
    p_index_for_distance = data['p_index_for_distances']
    
    p_index_for_angle = []
    p_index_for_angle.append([coor_transformer.ref_particle_2,
                              coor_transformer.ref_particle_1,
                              coor_transformer.ref_particle_3])
    for p in coor_transformer.particle_visited_in_order:
        p_index_for_angle.append(coor_transformer.angle_particle_idx[p])
    p_index_for_angle = torch.LongTensor(p_index_for_angle)

    p_index_for_dihedral = []
    for p in coor_transformer.particle_visited_in_order:
        p_index_for_dihedral.append(coor_transformer.dihedral_particle_idx[p])
    p_index_for_dihedral = torch.LongTensor(p_index_for_dihedral)

    num_particles = coor_transformer.num_particles
    nnforce = NNForce(num_particles,
                      hidden_size,
                      p_index_for_angle,
                      p_index_for_dihedral,
                      p_index_for_distance,
                      T)
    nnforce.nnforce_net.load_state_dict(state_dict)

    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)
    xyz = torch.from_numpy(traj_ref.xyz[0]).double()

    traced_nnforce = torch.jit.trace(nnforce, xyz)
    traced_nnforce.save(f"./output/{name}/FF/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_NNForceXYZ_weight_decay_{weight_decay:.3E}_different_initial_parameters.pth")

    nnforce = TorchForce(f"./output/{name}/FF/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_NNForceXYZ_weight_decay_{weight_decay:.3E}_different_initial_parameters.pth")
    nnforce.setForceGroup(0)
    system.addForce(nnforce)

    xml = omm.XmlSerializer.serialize(system)
    os.makedirs(f"./output/{name}/nnforce_system", exist_ok = True)
    with open(f"./output/{name}/nnforce_system/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_NNForce_weight_decay_{weight_decay:.3E}_different_initial_parameters.xml", 'w') as file_handle:
        file_handle.write(xml)

    fricCoef = 1./unit.picoseconds
    stepsize = 5. * unit.femtoseconds
    integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)

    platform = omm.Platform.getPlatformByName('Reference')
    context = omm.Context(system, integrator, platform)
    context.setPositions(xyz.numpy())
    state = context.getState(getEnergy = True)
    potential_energy = state.getPotentialEnergy()

    print(f"weight_decay: {weight_decay:.3E}", data['loss_wo_penalty'], potential_energy, flush = True)
exit()



import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'x-large'
mpl.rcParams['xtick.labelsize'] = 'x-large'
mpl.rcParams['ytick.labelsize'] = 'x-large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm

name = "2JOF"
include_rmsd = True
weight_decay = 2e-7

data = torch.load(f"./output/{name}/FF/FF_rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.pt")
FF = data['FF']
bonded_ff, lj_ff, rmsd_ff = FF['bonded'], FF['lj'], FF['rmsd']

fig = plt.figure()
fig.clf()
rmsd = np.linspace(rmsd_ff['min'].item(), rmsd_ff['max'].item(), rmsd_ff['U'].shape[1])
plt.plot(rmsd, rmsd_ff['U'][0,:].numpy(), linewidth = 2, color = 'black')
plt.savefig(f"./output/plots/{name}_u_rmsd.eps")




import mdtraj
import argparse
import math
import pickle
import matplotlib.pyplot as plt
import numpy as np
from sys import exit
import pandas as pd
import openmm.unit as unit

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
protein_info = pd.read_csv("./info/protein_temperature_and_ionic_strength.txt",
                           index_col = 'name', comment = '#')

for name in protein_names.iloc[0:2, 0]:
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
    traj_im = mdtraj.load_dcd(f"./output/{name}/traj_im/traj.dcd", psf)
    traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

    rmsd_md = mdtraj.rmsd(traj_md, traj_ref)
    rmsd_im = mdtraj.rmsd(traj_im, traj_ref)

    rmsd_max = max(rmsd_md.max(), rmsd_im.max())
    rmsd_max = min(2.0, rmsd_max)
    rmsd_min = 0.0

    width = 0.03
    size = int((rmsd_max - rmsd_min)/width) + 1
    
    rmsd_centers = np.linspace(rmsd_min, rmsd_max, size, endpoint = True)

    T = protein_info.loc[name, 'temperature'] * unit.kelvin
    kbT = unit.BOLTZMANN_CONSTANT_kB*unit.AVOGADRO_CONSTANT_NA*T
    kbT = kbT.value_in_unit(unit.kilojoule_per_mole)
    rmsd_k = 2.4/(width/1.5)**2

    print(f"name: {name}, rmsd_k: {rmsd_k:.2f}, size: {size}")
    
    with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'wb') as file_handle:
        pickle.dump({'size': size, 'centers': rmsd_centers, 'k': rmsd_k}, file_handle)



import mdtraj
import argparse
import math
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
import itertools
import pandas as pd

protein_names = pd.read_csv("./info/protein_names.txt",
                            comment = "#",
                            header = None)
protein_names = protein_names.iloc[0:2, 0].tolist()
flag_rmsd = [False, True]
weight_decay_list = [1e-7, 2e-7,
                     3e-7, 4e-7,
                     5e-7,
                     1e-6, 5e-6,
                     1e-5, 5e-5,
                     1e-4, 5e-4,
                     1e-3, 5e-3]
options = list(itertools.product(protein_names, flag_rmsd, weight_decay_list))

for name in protein_names:
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)    
    traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
    rmsd_md = mdtraj.rmsd(traj_md, traj_ref)

    for include_rmsd in flag_rmsd:
        pdf_hist =  PdfPages(f"./output/{name}/rmsd_full_rmsd_{include_rmsd}_hist.pdf")
        pdf_traj =  PdfPages(f"./output/{name}/rmsd_full_rmsd_{include_rmsd}_traj.pdf")

        for weight_decay in weight_decay_list:
            print(f"name: {name}, rmsd: {include_rmsd}, weight_decay: {weight_decay}", flush = True)
    
            traj_cg_NVT = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.dcd", psf)
            rmsd_cg_NVT = mdtraj.rmsd(traj_cg_NVT, traj_ref)        

            # traj_cg_TRE = mdtraj.load_dcd(f"./output/{name}/TRE/weight_decay_{weight_decay:.3E}/traj.dcd", psf)
            # rmsd_cg_TRE = mdtraj.rmsd(traj_cg_TRE, traj_ref)        

            fig = plt.figure()
            fig.clf()

            plt.hist(rmsd_md, bins = 40, alpha = 0.5, label = 'md', density = True)
            plt.hist(rmsd_cg_NVT, bins = 40, alpha = 0.5, label = 'cg_NVT', density = True)

            plt.legend()
            plt.title(f"weight_decay: {weight_decay:.3E}")
            pdf_hist.savefig()
            plt.close()

            fig = plt.figure(figsize = (6.4*4, 4.8))
            fig.clf()

            plt.plot(rmsd_md[::len(rmsd_md)//10000], alpha = 0.5, label = 'md')
            plt.plot(rmsd_cg_NVT[::len(rmsd_cg_NVT)//10000], alpha = 0.5, label = 'cg_NVT')

            plt.legend()
            plt.title(f"rmsd: {include_rmsd}, weight_decay: {weight_decay:.3E}")
            pdf_traj.savefig()
            plt.close()

        pdf_hist.close()
        pdf_traj.close()

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=NVT_LJ
#SBATCH --time=10:00:00
#SBATCH --partition=normal
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-8
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/NVT_LJ_%a.txt

import numpy as np
import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import argparse
import pandas as pd
import mdtraj
from sys import exit
import os
import time
import itertools
import torch
import pickle

name = "2JOF"
# include_rmsd = True
# weight_decay = 2e-7

include_rmsd = False
weight_decay = 4e-7

print(f"name: {name}, rmsd: {include_rmsd}, weight_decay: {weight_decay}", flush = True)

## read system
with open(f"./output/{name}/full_system/rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.xml", 'r') as file_handle:
    xml = file_handle.read()    
system = omm.XmlSerializer.deserialize(xml)

protein_info = pd.read_csv(
    "./info/protein_temperature_and_ionic_strength.txt",
    index_col = 'name',
    comment = '#')

## make an integrator
T = protein_info.loc[name, 'temperature']
kbT = unit.BOLTZMANN_CONSTANT_kB * T * unit.kelvin * unit.AVOGADRO_CONSTANT_NA

fricCoef = 1./unit.picoseconds
stepsize = 5. * unit.femtoseconds
integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)

## read topology and pdb file
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
#traj_full = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.dcd", psf)

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']
traj_imus = []
stride = 10
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj = traj[::stride]
    traj_imus.append(traj)
traj_imus = mdtraj.join(traj_imus)

traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)
xyz_init = traj_ref.xyz[0]

## minimize initial positions
platform = omm.Platform.getPlatformByName('Reference')
context = omm.Context(system, integrator, platform)

energy_md = []
for i in range(traj_md.n_frames):
    context.setPositions(traj_md.xyz[i])
    state = context.getState(getEnergy = True)
    potential_energy = state.getPotentialEnergy()
    energy_md.append(potential_energy / kbT)

    if (i + 1) % 10000 == 0:
        print(i)
        
energy_imus = []
for i in range(traj_imus.n_frames):
    context.setPositions(traj_imus.xyz[i])
    state = context.getState(getEnergy = True)
    potential_energy = state.getPotentialEnergy()
    energy_imus.append(potential_energy / kbT)

    if (i + 1) % 10000 == 0:
        print(i)
    
energy_md = torch.tensor(energy_md)
energy_imus = torch.tensor(energy_imus)

os.makedirs(f"./output/{name}", exist_ok = True)
torch.save({'md': energy_md,
            'imus': energy_imus},
           f"./output/{name}/full_energy_rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.pt")



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=elec
#SBATCH --time=03:00:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --mem=10G
#SBATCH --array=0-11
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/elec_%a.txt

import numpy as np
import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import argparse
import mdtraj
from sys import exit
import time
import pickle
import pandas as pd
import os
import torch

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

traj_md = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", psf, stride = 1)
with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']
traj_imus = []
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj_imus.append(traj)
traj_imus = mdtraj.join(traj_imus)
stride = size // 10
traj_imus = traj_imus[::stride]

with open(f"./output/{name}/system_im.xml", 'r') as file_handle:
    xml = file_handle.read()    
system = omm.XmlSerializer.deserialize(xml)

protein_info = pd.read_csv("./info/protein_temperature_and_ionic_strength.txt", index_col = 'name', comment = '#')
AA_info = pd.read_csv("./info/amino_acids_with_learned_sigmas.csv", index_col = 'name')

## make an integrator
T = protein_info.loc[name, 'temperature']
kbT = unit.BOLTZMANN_CONSTANT_kB * T * unit.kelvin * unit.AVOGADRO_CONSTANT_NA

def make_custom_elec_force():
    formula = ["ONE_4PI_EPS0*charge1*charge2/(r*epsilon)",
               "ONE_4PI_EPS0 = 138.935456",
               "epsilon = 78.4"
    ]
    custom_elec_force = omm.CustomNonbondedForce(";".join(formula))
    custom_elec_force.addPerParticleParameter('charge')
    return custom_elec_force

custom_elec_force = make_custom_elec_force()
custom_elec_force.setForceGroup(1)
for i in range(psf.n_residues):
    resname = psf.residue(i).name
    charge = AA_info.loc[resname, 'charge']
    custom_elec_force.addParticle([charge])
bonds = [[i, i+1] for i in range(psf.n_residues - 1)]
custom_elec_force.createExclusionsFromBonds(bonds, 3)    
system.addForce(custom_elec_force)

fricCoef = 1./unit.picoseconds
stepsize = 5. * unit.femtoseconds
integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)
platform = omm.Platform.getPlatformByName('Reference')
context = omm.Context(system, integrator, platform)

xyz_md = traj_md.xyz
elec_energy_md = []
for i in range(xyz_md.shape[0]):
    context.setPositions(xyz_md[i])
    state = context.getState(getEnergy = True, groups = set([1]))
    U = state.getPotentialEnergy()
    elec_energy_md.append(-U/kbT)
    
    if (i+1) % 10000 == 0:
        print(i)

xyz_imus = traj_imus.xyz
elec_energy_imus = []
for i in range(xyz_imus.shape[0]):
    context.setPositions(xyz_imus[i])
    state = context.getState(getEnergy = True, groups = set([1]))
    U = state.getPotentialEnergy()
    elec_energy_imus.append(-U/kbT)
    if (i+1) % 10000 == 0:
        print(i)
        
elec_energy_md = torch.tensor(elec_energy_md)
elec_energy_imus = torch.tensor(elec_energy_imus)

xml = omm.XmlSerializer.serialize(system)
f = open(f"./output/{name}/system_im_with_elec_simple.xml", 'w')
f.write(xml)
f.close()

os.makedirs(f"./output/{name}", exist_ok = True)
torch.save({'md': elec_energy_md,
            'imus': elec_energy_imus},
           f"./output/{name}/elec_energy_simple.pt")
exit()

# ONE_4PI_EPS0 = 138.935456
# EPS0 = 1./(ONE_4PI_EPS0*4*math.pi)
# kbT = unit.BOLTZMANN_CONSTANT_kB*300*unit.kelvin*unit.AVOGADRO_CONSTANT_NA
# kbT = kbT.value_in_unit(unit.kilojoule_per_mole)
# ionic_strength = 2/(40*unit.angstroms)**3
# ionic_strength = ionic_strength.value_in_unit(unit.nanometer**(-3))
# lambda_D = math.sqrt(EPS0*kbT/(2*ionic_strength))

# ionic_strength = protein_info.loc[name, 'ionic_strength']

# def make_custom_elec_force():
#     formula = ["ONE_4PI_EPS0*charge1*charge2/(r*epsilon)*exp(-r/lambda_D)",
#                "lambda_D = sqrt(EPS0*epsilon*kb*Temperature/(2*ionic_strength))",
#                "epsilon = A + B/(1 + kappa*exp(-lambda*B*r))",
#                "EPS0 = 1./(ONE_4PI_EPS0*4*PI)",
#                "PI = 3.141592653",
#                "B = 78.4 - A; A = -8.5525; kappa = 7.7839; lambda = 0.03627",
#                "kb = 2.4943387854459713/300",
#                "ONE_4PI_EPS0 = 138.935456"
#     ]

#     custom_elec_force = omm.CustomNonbondedForce(";".join(formula))
#     custom_elec_force.addGlobalParameter("Temperature", 300)
#     ionic_strength = 150*unit.millimolar
#     ionic_strength = ionic_strength*unit.AVOGADRO_CONSTANT_NA
#     ionic_strength = ionic_strength.value_in_unit(unit.nanometer**(-3))
#     custom_elec_force.addGlobalParameter("ionic_strength", ionic_strength)
#     custom_elec_force.addPerParticleParameter('charge')
#     print("remember to set values for global parameters T and ionic_strength")
#     return custom_elec_force



import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

score_md = pd.read_table('./output/test/TM_score_md.txt', header = None, sep = '[=(]')
score_md = score_md.iloc[:, 1].tolist()

score_cg = pd.read_table('./output/test/TM_score_cg.txt', header = None, sep = '[=(]')
score_cg = score_cg.iloc[:, 1].tolist()

fig = plt.figure()
fig.clf()
plt.hist(score_md, bins = 40, density = True, label = 'md', alpha = 0.5)
plt.hist(score_cg, bins = 40, density = True, label = 'cg', alpha = 0.5)
plt.savefig(f"./output/test/tmscore_hist.pdf")



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=basis_lj
#SBATCH --time=00:30:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --mem=100G
#SBATCH --array=0-1
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/basis_lj_%a.txt

import numpy as np
from scipy.interpolate import BSpline
from scipy.integrate import quad
import pickle
import mdtraj
from sys import exit
import os
import argparse
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
from CLCG.utils.splines import *
from CLCG.utils.CL import *
import torch
torch.set_default_dtype(torch.double)
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
mpl.rc('font', size = 16)
mpl.rc('axes', titlesize = 'large', labelsize = 'large')
mpl.rc('xtick', labelsize = 'large')
mpl.rc('ytick', labelsize = 'large')
import pandas as pd
import ray

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
job_idx = 1
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf, stride = 1)

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']
traj_imus = []
stride = 10
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj = traj[::stride]
    traj_imus.append(traj)
traj_imus = mdtraj.join(traj_imus)

indices = [(i,j) for i in range(psf.n_residues) for j in range(i+4, psf.n_residues)]
num_of_basis = 12

r_md = mdtraj.compute_distances(traj_md, indices)
r_md = torch.from_numpy(r_md).double()

r_min, _ = torch.min(r_md, 0)
r_max = 1.5

r_imus = mdtraj.compute_distances(traj_imus, indices)
r_imus = torch.from_numpy(r_imus).double()

ray.init(ignore_reinit_error = True, _temp_dir = "/home/gridsan/dingxq/tmp/ray")

r_md = ray.put(r_md)
r_imus = ray.put(r_imus)

r_grid_min = 0.2
r_grid_max = r_max
r_grid = ray.put(torch.linspace(r_grid_min, r_grid_max, 1000))

@ray.remote
def compute_lj_basis(j):
    import sys
    sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
    from CLCG.utils.splines import bs_lj
    print(j)
    basis_md = bs_lj(ray.get(r_md)[:,j], r_min[j].item(), r_max, num_of_basis, omega = False)
    basis_imus = bs_lj(ray.get(r_imus)[:,j], r_min[j].item(), r_max, num_of_basis, omega = False)
    basis_grid, omega = bs_lj(ray.get(r_grid), r_min[j].item(), r_max, num_of_basis, omega = True)
    
    return basis_md, basis_imus, basis_grid, omega

res = ray.get([compute_lj_basis.remote(j) for j in range(len(indices))])

basis_md = torch.cat([ res[i][0] for i in range(len(res))], dim = 1)
basis_imus = torch.cat([ res[i][1] for i in range(len(res))], dim = 1)
basis_grid = {}
basis_grid['basis'] = torch.stack([ res[i][2] for i in range(len(res))])
basis_grid['min'] = r_grid_min
basis_grid['max'] = r_grid_max
omega = torch.stack([ res[i][3] for i in range(len(res))])
basis_grid['omega'] = omega

# basis_md = basis_md.to_sparse()
# basis_imus = basis_imus.to_sparse()

torch.save({'md': basis_md,
            'imus': basis_imus,
            'grid': basis_grid},
           f"./output/{name}/basis_lj.pt")
exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=TRE
#SBATCH --time=3-00:00:00
#SBATCH --partition=xeon-p8
#SBATCH --cpus-per-task=25
#SBATCH --exclusive=user
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-27
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/TRE_%a.txt

import numpy as np
import openmm.app  as app
import openmm as omm
import openmm.unit as unit
import math
from sys import exit
import ray
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMRay")
from TRE import *
import argparse
import time
import pickle
import os
import pandas as pd
import mdtraj
import argparse

ray.init(num_cpus = int(os.environ['SLURM_CPUS_PER_TASK']),
         _temp_dir = '/home/gridsan/dingxq/tmp/ray')
print("available_resources: ", ray.available_resources())

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
weight_decay_list = [1e-8, 2e-8, 5e-8, 1e-7, 2e-7, 5e-7, 1e-6]

job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
protein_idx = job_idx // len(weight_decay_list)
weight_decay_idx = job_idx % len(weight_decay_list)

name = protein_names.iloc[protein_idx, 0]
weight_decay = weight_decay_list[weight_decay_idx]

print(f"name: {name}, weight_decay: {weight_decay}", flush = True)


## read system
with open(f"./output/{name}/full_system/weight_decay_{weight_decay:.3E}.xml", 'r') as file_handle:
    xml = file_handle.read()    
system = omm.XmlSerializer.deserialize(xml)

protein_info = pd.read_csv(
    "./info/protein_temperature_and_ionic_strength.txt",
    index_col = 'name',
    comment = '#')

## make an integrator
T = protein_info.loc[name, 'temperature']
fricCoef = 1./unit.picoseconds
stepsize = 5. * unit.femtoseconds
integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)

## read topology and pdb file
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)
xyz_init = traj_ref.xyz[0]

## minimize initial positions
platform = omm.Platform.getPlatformByName('Reference')
context = omm.Context(system, integrator, platform)
context.setPositions(xyz_init)

state = context.getState(getEnergy = True)
potential_energy = state.getPotentialEnergy()
print(potential_energy)

omm.LocalEnergyMinimizer_minimize(context)
state = context.getState(getEnergy = True,
                         getPositions = True)
potential_energy = state.getPotentialEnergy()
print(potential_energy)
init_positions = state.getPositions()

#### running TREMD
## construct TREActor
low_T = protein_info.loc[name, 'temperature']
high_T = 600
num_temperatures = int(os.environ['SLURM_CPUS_PER_TASK'])

temperatures =  1./np.linspace(1./low_T, 1./high_T, num_temperatures)
os.makedirs(f"./output/{name}/TRE/weight_decay_{weight_decay:.3E}", exist_ok = True)
dcd_file_names = [f"./output/{name}/TRE/weight_decay_{weight_decay:.3E}/traj_{i}.dcd"
                  for i in range(num_temperatures)]
actors = \
[ TREActor.options(num_cpus=1).remote(
    system, integrator, "Reference",
    temperatures[i], psf.to_openmm(),
    dcd_file_names[i], init_positions)
  for i in range(num_temperatures)]

## set all initial temperature to 100K
initial_T = 100.
[actor.update_temperature.remote(initial_T) for actor in actors]

## slowly increase temperature to the target tempearture for each replica
num_heating_steps = 125
dT = [(T - initial_T)/num_heating_steps for T in temperatures]
for idx_heating_step in range(num_heating_steps):
    print("idx_heating_step: {}".format(idx_heating_step))
    [actor.run_md.remote(1000) for actor in actors]
    [actors[i].update_temperature.remote(initial_T + (idx_heating_step + 1)*dT[i]) for i in range(len(actors))]
    
[actor.run_md.remote(10000) for actor in actors]

print("target temperatures:", temperatures)
print("final tempeartures:", ray.get([actor.get_temperature.remote() for actor in actors]))

start_time = time.time()

## run TREMD
num_steps = int(1e8)
exchange_freq = int(1e3)
save_freq = int(1e3)

# num_steps = int(2e5)
# exchange_freq = int(1e3)
# save_freq = int(1e3)

tre = TRE(actors)
tre.run(num_steps, exchange_freq, save_freq)

log = ray.get([actors[i].get_temperature_record.remote() for i in range(num_temperatures)])
log = np.array(log)

with open(f"./output/{name}/TRE/weight_decay_{weight_decay:.3E}/log_temperatures.pkl", 'wb') as file_handle:
    pickle.dump(log, file_handle)

print("used time: {:.2f}".format(time.time() - start_time))
ray.shutdown()    

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=make_system
#SBATCH --time=1:00:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-1
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/make_system_%a.txt

import numpy as np
import simtk.openmm as omm
import simtk.unit as unit
import simtk.openmm.app as ommapp
import math
import mdtraj
import pickle
from sys import exit
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG/")
from CLCG.utils.make_system import *
import argparse
import pandas as pd
import os

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

protein_info = pd.read_csv("./info/protein_temperature_and_ionic_strength.txt",
                           index_col = 'name', comment = '#')
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
AA_info = pd.read_csv("./info/amino_acids_with_learned_sigmas.csv", index_col = 'name')
masses = [AA_info.loc[r.name, 'mass'] for r in psf.residues]

with open(f"./output/{name}/bonded_parameters_im.pkl", 'rb') as file_handle:
    bonded_parameters = pickle.load(file_handle)

with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
    coor_transformer = pickle.load(file_handle)

system = make_system(masses,
                     coor_transformer,
                     bonded_parameters,
                     protein_info.loc[name, 'temperature'],
                     jacobian = True)

xml = omm.XmlSerializer.serialize(system)
f = open(f"./output/{name}/system_im.xml", 'w')
f.write(xml)
f.close()

exit()
